"""
TODO
"""
from typing import List, Optional

import matplotlib.pyplot as plt
import numpy as _np
import pandas as pd
import scipy
from numpy.polynomial import Polynomial
from scipy.interpolate import interp1d
from scipy.optimize import Bounds, minimize, newton_krylov


def compute_dvh(dose_data, voxel_volume):
    """
    Args:
        dose_data: 3D-array of the dose values.
        voxel_volume: The volume of a unique voxel. We consider the same size for all voxels.

    Returns:
        dvh_dataframe: The dvh stored into a 2 columns dataframe

    """
    dvh_histogram = plt.hist(
        dose_data.flatten(),
        bins=100,
        cumulative=-1,
        density=True,
    )

    dvh_dataframe = pd.DataFrame(
        columns=[
            "volume",
            "dose_value",
        ],
    )

    dvh_dataframe["volume"] = dvh_histogram[0]
    dvh_dataframe["dose_value"] = voxel_volume * (
        dvh_histogram[1][:-1] + (dvh_histogram[1][1] - dvh_histogram[1][0]) / 2
    )

    return dvh_dataframe


class BraggPeakException(Exception):
    def __init__(self, m: str = "") -> None:  # pragma: no cover
        self.message = m


class BraggPeakAnalysis:
    """
    This class processes a normalized 1D depth dose profile, or a pristine Bragg peak, to extract key information such
    as the maximum dose position, the distal range at a given percentage of the maximum, and the distal fall-off (DFO).
    The user must provide both the dose data and the corresponding positions in depth.
    """

    def __init__(self, bp: pd.DataFrame, method: str, low_dose=1, high_dose=70):
        self.data = bp
        self.low_bound = low_dose
        self.high_bound = high_dose
        self.coefficients = None
        self.bp_interval = None
        self.method = method

    def get_distal_interval(self):
        idx = _np.where(_np.isclose(self.data[self.data.columns[1]], 1e2))
        distal_interval = self.data[idx[0][0] :]

        return distal_interval

    def get_bp_interval(self):
        idx1 = _np.where(self.data[self.data.columns[1]] > self.high_bound)
        idx2 = _np.where(self.data[self.data.columns[1]] < self.low_bound)

        idx_min = idx1[0][0]
        idx_max = idx2[0][0]
        bp_interval = self.data[idx_min:idx_max]

        return bp_interval

    def compute_percentage(self, x):  # Gives dose percentage in the distal region for a given z position
        f = interp1d(
            self.get_distal_interval()[self.get_distal_interval().columns[0]].values,
            self.get_distal_interval()[self.get_distal_interval().columns[1]].values,
            kind=2,
            bounds_error=False,
        )

        return f(x)

    def fitting_function(self, x):
        coeff = self.coefficients.convert().coef
        return (
            coeff[0]
            + coeff[1] * x
            + coeff[2] * x**2
            + coeff[3] * x**3
            + coeff[4] * x**4
            + coeff[5] * x**5
            + coeff[6] * x**6
            + coeff[7] * x**7
            + coeff[8] * x**8
        )

    def fit_bp(self):
        """
        This function fits a Bragg peak using a 8th order polynom. Based on reference xxx.

        Args:
        Returns:

        """

        bp_interval = self.get_bp_interval()
        bp_interval.reset_index(drop=True, inplace=True)

        x = _np.float32(bp_interval[bp_interval.columns[0]])
        y = _np.float32(bp_interval[bp_interval.columns[1]])

        self.coefficients = Polynomial.fit(
            x=x,
            y=y,
            deg=8,
            full=False,
        )

        self.bp_interval = bp_interval[bp_interval.columns[0]].to_numpy()

        return self.bp_interval, self.coefficients, self.fitting_function(self.bp_interval)

    def get_coefficients(self) -> _np.array:
        return self.coefficients.convert().coef

    def get_xrange(self, val):
        if self.method == "polynomial_fit":
            rval = (self.fit_bp()[1] - val).roots()
            real_val = rval[rval.imag == 0].real
            var = real_val[
                _np.where(
                    _np.logical_and(
                        real_val >= self.get_distal_interval()[self.get_distal_interval().columns[0]].values[0],
                        real_val <= self.get_distal_interval()[self.get_distal_interval().columns[0]].values[-1],
                    ),
                )
            ]
            # check if we are in the distal interval of the BP
            if var.size:
                return _np.max(var)
            return 0

        elif self.method == "scipy.optimize":
            nearest_val = self.get_distal_interval()[self.get_distal_interval().columns[0]][
                _np.abs(self.get_distal_interval()[self.get_distal_interval().columns[1]] - val).idxmin()
            ]

            var = newton_krylov(
                lambda x: _np.abs(self.compute_percentage(x=x) - val),
                xin=nearest_val,
                maxiter=10000,
                f_tol=1e-12,
                x_tol=1e-13,
            )

            return var

        else:
            raise BraggPeakException(
                "Please choose a valid computation method (must be either 'polynomial_fit' or 'scipy.optimize')",
            )

    def get_r90(self) -> float:
        return self.get_xrange(90)

    def get_r80(self) -> float:
        return self.get_xrange(80)

    def get_r20(self) -> float:
        return self.get_xrange(20)

    def get_r10(self) -> float:
        return self.get_xrange(10)

    def get_maximum(self) -> float:
        roots = self.fit_bp()[1].deriv().roots()
        real_roots = roots[roots.imag == 0].real
        var = real_roots[
            _np.where(
                _np.logical_and(
                    real_roots >= self.get_distal_interval()[self.get_distal_interval().columns[0]].values[0],
                    real_roots <= self.get_distal_interval()[self.get_distal_interval().columns[0]].values[-1],
                ),
            )
        ]
        fit_var = self.fitting_function(var)
        var = var[_np.where(fit_var == _np.max(fit_var))]  # take the value where the fit is maximum (global maximum)
        return var[0]

    def compute_distal_fall_off(self) -> List[float]:
        return [self.get_r10() - self.get_r90(), self.get_r20() - self.get_r80()]

    def compute_dfo_90_10(self) -> float:
        return self.compute_distal_fall_off()[0]

    def compute_dfo_80_20(self) -> float:
        return self.compute_distal_fall_off()[1]


class SpreadOutBraggPeakAnalysis:
    """
    This class takes a set of Bragg peaks, or a Bragg peak library, as input and computes the relative weights required
    to obtain a uniform depth dose profile, known as the SOBP. Additionally, this class processes the resulting SOBP
    and provides the flatness, DFO, and distal range at a given percentage of the maximum to the user.
    """

    def __init__(
        self,
        dose_data: Optional[pd.DataFrame] = None,
        method: str = scipy.optimize,
        z_axis: Optional[_np.array] = None,
        color: str = "blue",
        str_on_legend: str = "",
        adjust_last_peak: float = 1.0,
    ):
        self.dose_data = dose_data
        self.method = method
        self.z_axis = z_axis
        self.modul_type = "Full"
        self.color = color
        self.str_on_legend = str_on_legend
        self.adjust_last_peak = adjust_last_peak

    def get_library_max_ranges(self):
        max_ranges = _np.zeros(self.dose_data.shape[0])

        for i in range(self.dose_data.shape[0]):
            normalized_bragg_peak = 1e2 * self.dose_data.iloc[i, :] / _np.max(self.dose_data.iloc[i, :])
            bp_analysis = BraggPeakAnalysis(
                bp=pd.DataFrame(
                    {
                        "centers": self.z_axis,
                        "dose": normalized_bragg_peak,
                    },
                ),
                method="scipy.optimize",
            )

            max_ranges[i] = bp_analysis.get_xrange(100)

        return max_ranges

    def sobp_data(self) -> _np.array:
        sobp_data = _np.zeros(
            (
                self.dose_data.shape[0],
                self.dose_data.shape[0],
            ),
        )  # Array to store maximum dose values of each Bragg Peak

        max_ranges = self.get_library_max_ranges()
        for i in range(self.dose_data.shape[0]):
            for j in range(self.dose_data.shape[0]):
                bp_analysis = BraggPeakAnalysis(
                    bp=pd.DataFrame(
                        {
                            "z": self.z_axis,
                            "dose": self.dose_data.iloc[j, :],
                        },
                    ),
                    method="scipy.optimize",
                )
                f = interp1d(
                    bp_analysis.data[bp_analysis.data.columns[0]].values,
                    bp_analysis.data[bp_analysis.data.columns[1]].values,
                    kind=2,
                    bounds_error=False,
                )
                sobp_data[i, j] = f(max_ranges[i])

        return sobp_data

    def compute_weights(self) -> _np.array:
        a_matrix = self.sobp_data() / self.sobp_data().max()  # Normalize all data to max 1
        goal_dose_values = _np.full(a_matrix.shape[0], 1)

        if self.method == "np.linalg.solve":  # This method may give negative weights !
            return _np.linalg.solve(a_matrix, goal_dose_values)

        elif self.method == "scipy.optimize":
            b = goal_dose_values
            n = len(b)
            fun = lambda x: _np.linalg.norm(_np.dot(a_matrix, x) - b)  # noqa: E731
            sol = minimize(
                fun,
                _np.zeros(n),
                method="L-BFGS-B",
                bounds=[(0.0, None) for _ in range(n)],
            )

            return sol["x"]

    def get_final_weights(self):
        if self.adjust_last_peak is not None:
            weights = self.compute_weights()
            weights[0] *= self.adjust_last_peak

        return weights

    def compute_sobp_profile(self) -> _np.array:
        weights = self.get_final_weights()

        weights = weights.reshape(1, self.dose_data.shape[0])
        weighted_dose_data = _np.matmul(
            weights,
            self.dose_data.values,
        )

        sobp_profile = weighted_dose_data.reshape(self.dose_data.shape[1])

        return 1e2 * sobp_profile / sobp_profile.max()

    def view_sobp(self, with_pristine_peaks=False):
        def plot(x, y):
            plt.plot(
                x,
                y,
                linestyle="dashed",
                linewidth=0.7,
                color="r",
                marker="*",
                markersize=4,
            )

        if with_pristine_peaks:
            weighted_dose = self.dose_data.multiply(self.compute_weights(), axis=0)
            weighted_dose.apply(
                lambda e: plot(
                    x=self.z_axis,
                    y=1e2 * e.values / self.sobp_data().max(),
                ),
                axis="columns",
            )

        plt.plot(
            self.z_axis,
            self.compute_sobp_profile(),
            linestyle="dashed",
            linewidth=0.8,
            color=self.color,
            marker="*",
            markersize=5,
            label=self.str_on_legend,
        )

        plt.xlabel("Depth in water (mm)")
        plt.ylabel("Normalized dose (\\%)")
        plt.legend(loc=(0.15, 1.01), fontsize=20, ncol=2, frameon=False)
        plt.xlim(0, _np.max(self.z_axis))

    def recompute_sobp_for_another_range(self, bp_to_leave):
        if bp_to_leave == 0:
            new_weighted_dose = _np.matmul(
                self.compute_weights().reshape(1, self.dose_data.shape[0]),
                self.dose_data.values[0:, :],
            )
        else:
            new_shape = self.dose_data.shape[0] - bp_to_leave
            weights = self.compute_weights()[0:new_shape]

            new_weighted_dose = _np.matmul(
                weights.reshape(1, new_shape),
                self.dose_data.values[bp_to_leave:, :],
            )

        sobp_data = new_weighted_dose.reshape(self.dose_data.shape[1])
        sobp_data = 1e2 * sobp_data / sobp_data.max()

        return sobp_data

    def recompute_sobp_for_another_width(self, bp_to_leave):
        new_shape = self.dose_data.shape[0] - bp_to_leave
        weights = self.compute_weights()[0:new_shape]
        new_weighted_dose = _np.matmul(
            weights.reshape(1, new_shape),
            self.dose_data.values[0:new_shape, :],
        )

        sobp_data = new_weighted_dose.reshape(self.dose_data.shape[1])
        sobp_data = 1e2 * sobp_data / sobp_data.max()

        return sobp_data

    def compute_ranges(self):
        sobp_array = self.compute_sobp_profile()
        idxmax = _np.where(sobp_array < 1)[0][0]  # Select the dose values higher than 1%

        idx = 0
        for i in range(_np.flip(sobp_array[0:idxmax]).shape[0]):
            if _np.flip(sobp_array[0:idxmax])[i + 1] > _np.flip(sobp_array[0:idxmax])[i]:
                idx += 1
            else:
                break

        flipped_data = _np.flip(sobp_array[idxmax - idx : idxmax])  # Select the ramp

        def compute_range(data, r, z_axis):
            idx_xin = pd.DataFrame(_np.abs(data - r)).idxmin()
            nearest_val = _np.flip(z_axis[idxmax - idx : idxmax])[idx_xin[0]]
            fct = interp1d(
                _np.flip(z_axis[idxmax - idx : idxmax]),
                data,
                kind=2,
                bounds_error=False,
            )

            res = minimize(
                fun=lambda x: _np.abs(fct(x=x) - r),
                bounds=Bounds(
                    lb=nearest_val - 0.5,
                    ub=nearest_val + 0.5,
                ),
                x0=nearest_val,
                method="trust-constr",
                options={
                    "xtol": 1e-8,
                    "verbose": 0,
                    "maxiter": 1e5,
                },
            )
            return res

        r_10 = compute_range(flipped_data, 10, self.z_axis).x[0]
        r_20 = compute_range(flipped_data, 20, self.z_axis).x[0]
        r_80 = compute_range(flipped_data, 80, self.z_axis).x[0]
        r_90 = compute_range(flipped_data, 90, self.z_axis).x[0]

        return r_10, r_20, r_80, r_90

    def get_sobp_r10(self):
        return self.compute_ranges()[0]

    def get_sobp_r20(self):
        return self.compute_ranges()[1]

    def get_sobp_r80(self):
        return self.compute_ranges()[2]

    def get_sobp_r90(self):
        return self.compute_ranges()[3]

    def get_sobp_flatness(self):
        sobp_array = self.compute_sobp_profile()
        idx_98 = _np.where(_np.isclose(sobp_array, 98, atol=1.5))[0]
        max_dose = _np.max(sobp_array[idx_98[0] : idx_98[-1]])
        min_dose = _np.min(sobp_array[idx_98[0] : idx_98[-1]])

        flatness = 1e2 * (max_dose - min_dose) / (max_dose + min_dose)

        return flatness

    def get_ur_min_dose(self):
        sobp_array = self.compute_sobp_profile()
        idx_98 = _np.where(_np.isclose(sobp_array, 98, atol=1.5))[0]
        min_dose = _np.min(sobp_array[idx_98[0] : idx_98[-1]])

        return min_dose

    def get_ur_max_dose(self):
        sobp_array = self.compute_sobp_profile()
        idx_98 = _np.where(_np.isclose(sobp_array, 98, atol=1.5))[0]
        max_dose = _np.max(sobp_array[idx_98[0] : idx_98[-1]])

        return max_dose


class LateralProfileAnalysis:
    """
    This class takes a normalized 1D transverse dose profile as input and computes various parameters such as the field
    size, the uniform region (defined as 80% of the field size), the transverse flatness of the uniform region, and the
    lateral penumbra at the left and right of this region. The user must provide both the dose data and the
    corresponding transverse positions.
    """

    def __init__(self, dose_profile: _np.array, positions: _np.array):
        self.dose_profile = dose_profile
        self.positions = positions

    def set_data(self):
        idxmax = _np.where(self.positions == self.positions[_np.int(self.positions.shape[0] / 2)])[0][0]
        positions_left = self.positions[0:idxmax]
        positions_right = self.positions[idxmax:]
        dose_left = self.dose_profile[0:idxmax]
        dose_right = self.dose_profile[idxmax:]

        return idxmax, positions_left, dose_left, positions_right, dose_right

    def define_f(self):
        f_left = interp1d(
            self.set_data()[1],
            self.set_data()[2],
            kind=1,
            bounds_error=False,
        )
        f_right = interp1d(
            self.set_data()[3],
            self.set_data()[4],
            kind=1,
            bounds_error=False,
        )
        return f_left, f_right

    def get_position_left(self, percentage):
        nearest_val = self.set_data()[1][pd.DataFrame(_np.abs(self.set_data()[2] - percentage)).idxmin()]
        position = minimize(
            fun=lambda x: _np.abs(self.define_f()[0](x=x) - percentage),
            bounds=Bounds(
                lb=nearest_val - 3,
                ub=nearest_val + 3,
            ),
            x0=nearest_val,
            method="trust-constr",
            options={
                "xtol": 1e-8,
                "verbose": 0,
                "maxiter": 1e5,
            },
        )
        return position.x[0]

    def get_position_right(self, percentage):
        nearest_val = self.set_data()[3][pd.DataFrame(_np.abs(self.set_data()[4] - percentage)).idxmin()]
        position = minimize(
            fun=lambda x: _np.abs(self.define_f()[1](x=x) - percentage),
            bounds=Bounds(
                lb=nearest_val - 3,
                ub=nearest_val + 3,
            ),
            x0=nearest_val,
            method="trust-constr",
            options={
                "xtol": 1e-8,
                "verbose": 0,
                "maxiter": 1e5,
            },
        )
        return position.x[0]

    def get_p_20_left(self):
        return self.get_position_left(20)

    def get_p_50_left(self):
        return self.get_position_left(50)

    def get_p_80_left(self):
        return self.get_position_left(80)

    def get_p_20_right(self):
        return self.get_position_right(20)

    def get_p_50_right(self):
        return self.get_position_right(50)

    def get_p_80_right(self):
        return self.get_position_right(80)

    def get_field_size(self):
        return self.get_p_50_right() - self.get_p_50_left()

    def get_ur_left(self):
        return self.get_p_50_left() + 2 * self.get_penumbra_left()

    def get_ur_right(self):
        return self.get_p_50_right() - 2 * self.get_penumbra_right()

    def get_ur_size(self):
        return self.get_ur_right() - self.get_ur_left()

    def get_ur_min_dose(self):
        ur_idx_min = _np.where(self.positions > self.get_ur_left())[0][0]
        ur_idx_max = _np.where(self.positions > self.get_ur_right())[0][0]

        return _np.min(self.dose_profile[ur_idx_min:ur_idx_max])

    def get_ur_max_dose(self):
        ur_idx_min = _np.where(self.positions > self.get_ur_left())[0][0]
        ur_idx_max = _np.where(self.positions > self.get_ur_right())[0][0]

        return _np.max(self.dose_profile[ur_idx_min:ur_idx_max])

    def get_ur_flatness(self):
        return (
            1e2 * (self.get_ur_max_dose() - self.get_ur_min_dose()) / (self.get_ur_max_dose() + self.get_ur_min_dose())
        )

    def get_penumbra_left(self):
        return self.get_p_80_left() - self.get_p_20_left()

    def get_penumbra_right(self):
        return self.get_p_20_right() - self.get_p_80_right()

    def get_penumbra(self):
        return _np.mean([self.get_penumbra_left(), self.get_penumbra_right()])

    def plot(self):
        plt.vlines(
            x=[
                self.get_p_20_left(),
                self.get_p_50_left(),
                self.get_p_80_left(),
                self.get_p_20_right(),
                self.get_p_50_right(),
                self.get_p_80_right(),
            ],
            ymin=0,
            ymax=100,
            color="k",
            linestyle="dashed",
        )

        plt.hlines(
            y=[100, self.get_ur_min_dose(), self.get_ur_max_dose()],
            xmin=self.get_p_20_left(),
            xmax=self.get_p_20_right(),
            color="k",
            linestyle="dashed",
        )

        plt.text(
            s=f"Flatness: {_np.round(self.get_ur_flatness(), 2)} \\%",
            x=-self.get_field_size() / 5,
            y=85,
            fontsize=15,
            color="b",
        )
        plt.text(
            s=f"P$_L$: {_np.round(self.get_penumbra_left(), 2)} mm",
            x=self.get_p_50_left() - 2 * self.get_penumbra_left(),
            y=30,
            fontsize=15,
            color="b",
        )
        plt.text(
            s=f"P$_R$: {_np.round(self.get_penumbra_right(), 2)} mm",
            x=self.get_p_50_right() + self.get_penumbra_right(),
            y=30,
            fontsize=15,
            color="b",
        )

        plt.plot(
            self.positions,
            self.dose_profile,
            marker="o",
            markersize=1.5,
            linestyle="dashed",
            color="b",
        )

        plt.xlabel("Transverse position (mm)")
        plt.ylabel("Normalized dose (\\%)")

        plt.grid()


class GammaAnalysis:
    pass


class RegularSpotScanning:
    """
    This class provides a method to calculate the required spot spacing for a regular grid irradiation scheme in order
    to obtain a two-dimensional uniform dose deposition profile for a given spot width (1 sigma) and a targeted
    circular field. The user inputs the half value of the field size, the desired number of spots along each axis of
    the field, and the standard deviation (1 sigma) of the beam. The required spot spacing to achieve a 2D dose
    uniformity of at least 98% is directly outputted.
    """

    def __init__(self, sigma, fieldsize, n_spots_per_axis):
        self.sigma = sigma
        self.fieldsize = fieldsize
        self.n_spots_per_axis = n_spots_per_axis

    def double_gaussian_function(self, x, y, a, mu_x, mu_y):
        return (
            a
            * _np.exp(
                -((x - mu_x) ** 2) / (2 * self.sigma**2) - (y - mu_y) ** 2 / (2 * self.sigma**2),
            )
            / (_np.sqrt(2 * _np.pi) * self.sigma)
        )

    def compute_2d_scanned_profile(self, spot_space):
        positions = _np.arange(-50, 51, 1)
        x, y = _np.meshgrid(positions, positions)
        shape = x.shape

        dose_df = pd.DataFrame(
            {
                "x": x.reshape(shape[0] * shape[1]),
                "y": y.reshape(shape[0] * shape[1]),
            },
        )

        for i in range(self.n_spots_per_axis):
            for j in range(self.n_spots_per_axis):
                dose_df[f"dose_{i}_{j}"] = self.double_gaussian_function(
                    x=dose_df["x"],
                    y=dose_df["y"],
                    a=100,
                    mu_x=(-(self.n_spots_per_axis - 1) / 2 + i) * spot_space,
                    mu_y=(-(self.n_spots_per_axis - 1) / 2 + j) * spot_space,
                )

        total_dose = dose_df.drop(columns=["x", "y"]).sum(axis=1)
        dose_df["total_dose"] = total_dose

        return dose_df

    def optimize_regular_grid_spots_placement(self):
        def compute_2d_profile_flatness(x):
            df = self.compute_2d_scanned_profile(spot_space=x)
            profile = df.query(f"x**2+y**2<={self.fieldsize**2}")["total_dose"]

            flatness = 1e2 * (_np.max(profile) - _np.min(profile)) / (_np.max(profile) + _np.min(profile))

            return flatness

        optim = minimize(
            fun=compute_2d_profile_flatness,
            bounds=Bounds(
                lb=0,
                ub=5 * self.sigma,
            ),
            x0=3 * self.sigma,
            method="trust-constr",
            options={
                "verbose": 1,
                "xtol": 1e-20,
                "maxiter": 1e6,
            },
        )

        return optim


class ContourSpotScanning:
    """
    This class works similarly to RegularSpotScanning but uses a circular, contour-based irradiation scheme with a
    central spot placed at the center of the field. The user can choose whether to impose the irradiation radius. The
    output of the calculation includes the relative weight of the contour spots compared to the central spot and the
    angle spacing between these spots.
    """

    def __init__(self, sigma, fieldsize, desired_angle, shoot_on_aperture: bool = True, angle_imposed: bool = False):
        self.sigma = sigma
        self.fieldsize = fieldsize
        self.desired_angle = desired_angle
        self.shoot_on_aperture = shoot_on_aperture
        self.angle_imposed = angle_imposed

    def double_gaussian_function(self, x, y, a, mu_x, mu_y):
        return (
            a
            * _np.exp(
                -((x - mu_x) ** 2) / (2 * self.sigma**2) - (y - mu_y) ** 2 / (2 * self.sigma**2),
            )
            / (_np.sqrt(2 * _np.pi) * self.sigma)
        )

    def compute_2d_contour_scanned_profile(self, angle, weight):
        positions = _np.arange(-50, 51, 1)
        x, y = _np.meshgrid(
            positions,
            positions,
        )

        n_contour_spots = _np.int(2 * _np.pi / angle)
        dose_df = pd.DataFrame(
            {
                "x": x.reshape(x.shape[0] * x.shape[1]),
                "y": y.reshape(x.shape[0] * x.shape[1]),
            },
        )

        dose_df["dose_central_spot"] = self.double_gaussian_function(
            x=dose_df["x"],
            y=dose_df["y"],
            a=100,
            mu_x=0.0,
            mu_y=0.0,
        )
        for i in range(n_contour_spots):
            tmp_theta = -_np.pi + i * angle
            dose_df[f"dose_{i}"] = (
                self.double_gaussian_function(
                    x=dose_df["x"],
                    y=dose_df["y"],
                    a=100,
                    mu_x=self.fieldsize * _np.cos(tmp_theta),
                    mu_y=self.fieldsize * _np.sin(tmp_theta),
                )
                * weight
            )

        total_dose = dose_df.drop(columns=["x", "y"]).sum(axis=1)
        dose_df["total_dose"] = total_dose

        return dose_df

    def compute_2d_contour_scanned_profile_angle_imposed(self, weight):
        positions = _np.arange(-50, 51, 1)
        x, y = _np.meshgrid(
            positions,
            positions,
        )

        n_contour_spots = _np.int(2 * _np.pi / self.desired_angle)
        dose_df = pd.DataFrame(
            {
                "x": x.reshape(x.shape[0] * x.shape[1]),
                "y": y.reshape(x.shape[0] * x.shape[1]),
            },
        )

        dose_df["dose_central_spot"] = self.double_gaussian_function(
            x=dose_df["x"],
            y=dose_df["y"],
            a=100,
            mu_x=0.0,
            mu_y=0.0,
        )
        for i in range(n_contour_spots):
            tmp_theta = -_np.pi + i * self.desired_angle
            dose_df[f"dose_{i}"] = (
                self.double_gaussian_function(
                    x=dose_df["x"],
                    y=dose_df["y"],
                    a=100,
                    mu_x=self.fieldsize * _np.cos(tmp_theta),
                    mu_y=self.fieldsize * _np.sin(tmp_theta),
                )
                * weight
            )

        total_dose = dose_df.drop(columns=["x", "y"]).sum(axis=1)
        dose_df["total_dose"] = total_dose

        return dose_df

    def compute_2d_contour_scanned_profile_radius_not_imposed(self, angle, weight, shoot_at):
        positions = _np.arange(-50, 51, 1)
        x, y = _np.meshgrid(
            positions,
            positions,
        )

        n_contour_spots = _np.int(2 * _np.pi / angle)
        dose_df = pd.DataFrame(
            {
                "x": x.reshape(x.shape[0] * x.shape[1]),
                "y": y.reshape(x.shape[0] * x.shape[1]),
            },
        )

        dose_df["dose_central_spot"] = self.double_gaussian_function(
            x=dose_df["x"],
            y=dose_df["y"],
            a=100,
            mu_x=0.0,
            mu_y=0.0,
        )
        for i in range(n_contour_spots):
            tmp_theta = -_np.pi + i * angle
            dose_df[f"dose_{i}"] = (
                self.double_gaussian_function(
                    x=dose_df["x"],
                    y=dose_df["y"],
                    a=100,
                    mu_x=shoot_at * _np.cos(tmp_theta),
                    mu_y=shoot_at * _np.sin(tmp_theta),
                )
                * weight
            )

        total_dose = dose_df.drop(columns=["x", "y"]).sum(axis=1)
        dose_df["total_dose"] = total_dose

        return dose_df

    def optimize_contour_spots(self):
        def compute_contour_profile_flatness(x):
            df = self.compute_2d_contour_scanned_profile(
                angle=x[0],
                weight=x[1],
            )
            profile = df.query("x**2+y**2<=@self.fieldsize**2")["total_dose"]
            flatness = 1e2 * (_np.max(profile) - _np.min(profile)) / (_np.max(profile) + _np.min(profile))

            return flatness

        def compute_contour_profile_flatness_angle_imposed(x):
            df = self.compute_2d_contour_scanned_profile_angle_imposed(weight=x)
            profile = df.query("x**2+y**2<=@self.fieldsize**2")["total_dose"]
            flatness = 1e2 * (_np.max(profile) - _np.min(profile)) / (_np.max(profile) + _np.min(profile))

            return flatness

        def compute_contour_profile_flatness_radius_not_imposed(x):
            df = self.compute_2d_contour_scanned_profile_radius_not_imposed(
                angle=x[0],
                weight=x[1],
                shoot_at=x[2],
            )
            profile = df.query("x**2+y**2<=@self.fieldsize**2")["total_dose"]
            flatness = 1e2 * (_np.max(profile) - _np.min(profile)) / (_np.max(profile) + _np.min(profile))

            return flatness

        if self.shoot_on_aperture and not self.angle_imposed:
            optim = minimize(
                fun=compute_contour_profile_flatness,
                bounds=Bounds(
                    [0, 0],
                    [_np.pi, 10],
                ),
                x0=[_np.pi / 6, 0.5],
                method="trust-constr",
                options={"verbose": 1, "xtol": 1e-10, "maxiter": 1e5},
            )

        elif self.shoot_on_aperture and self.angle_imposed:
            optim = minimize(
                fun=compute_contour_profile_flatness_angle_imposed,
                bounds=Bounds(
                    [0],
                    [10],
                ),
                x0=0.5,
                method="trust-constr",
                options={"verbose": 1, "xtol": 1e-10, "maxiter": 1e5},
            )

        else:
            optim = minimize(
                fun=compute_contour_profile_flatness_radius_not_imposed,
                bounds=Bounds(
                    [0, 0, 0],
                    [_np.pi, 10, 2 * self.fieldsize],
                ),
                x0=[_np.pi / 6, 0.5, self.fieldsize],
                method="trust-constr",
                options={"verbose": 1, "xtol": 1e-10, "maxiter": 1e5},
            )

        return optim



"""
TODO
"""
from typing import Mapping, Optional

import pandas as _pd
from georges_core.sequences import Sequence

from .. import ureg as _ureg
from .mcs import DifferentialMoliere as _DifferentialMoliere


def track_energy(sequence: Sequence, energy: _ureg.Quantity):
    s: _pd.DataFrame = sequence.to_df()
    energy = energy.m_as("MeV")
    if "MATERIAL" not in s.columns:
        s["ENERGY_IN"] = energy
        s["ENERGY_OUT"] = energy
        return s
    for i, e in s.iterrows():
        s.loc[i, "ENERGY_IN"] = energy
        if not _pd.isnull(e["MATERIAL"]) and e["L"] != 0.0:
            e = e["MATERIAL"].stopping(thickness=e["L"], kinetic_energy=energy * _ureg.MeV)
            if e is not None:
                energy = e.ekin.m_as("MeV")
        s.loc[i, "ENERGY_OUT"] = energy
        s.loc[i, "DELTA_E"] = s.at[i, "ENERGY_IN"] - energy
    return s


def propagate(sequence: Sequence, energy: _ureg.Quantity, beam: Optional[Mapping] = None, model=_DifferentialMoliere):
    """

    Args:
        sequence: Sequence to track in
        energy: Beam energy
        beam: A_i beam parameters
        model: Model to use. Default is DifferentialMoliere

    Returns:

    """
    # Default beam
    if beam is None:
        beam = {
            "A0": 0,
            "A1": 0,
            "A2": 0,
        }

    # Compute energy loss along the line
    s = track_energy(sequence=sequence, energy=energy)

    # Beam spreading and scattering following the Fermi-Eyges model
    for i, e in s.iterrows():
        if not _pd.isnull(e["MATERIAL"]):
            fe = e["MATERIAL"].scattering(
                kinetic_energy=energy,
                thickness=e["L"],
                model=model,
            )
        else:
            fe = {"A": (0, 0, 0), "B": 0}
        s.at[i, "A0"] = fe["A"][0]
        s.at[i, "A1"] = fe["A"][1]
        s.at[i, "A2"] = fe["A"][2]
        s.at[i, "B"] = fe["B"]

    # Sum the matrix elements
    acc = [beam["A0"], beam["A1"], beam["A2"]]
    for i, e in s.iterrows():
        s.loc[i, "A0_IN"] = acc[0]
        s.loc[i, "A1_IN"] = acc[1]
        s.loc[i, "A2_IN"] = acc[2]
        s.loc[i, "A0_OUT"] = acc[0] + e["A0"]
        s.loc[i, "A1_OUT"] = acc[0] * e["L"].m_as("m") + acc[1] + e["A1"]
        s.loc[i, "A2_OUT"] = acc[0] * e["L"].m_as("m") ** 2 + 2 * acc[1] * e["L"].m_as("m") + acc[2] + e["A2"]
        acc = [s.loc[i, "A0_OUT"], s.loc[i, "A1_OUT"], s.loc[i, "A2_OUT"]]
    s["B_IN"] = s["A0_IN"] * s["A2_IN"] - s["A1_IN"] ** 2
    s["B_OUT"] = s["A0_OUT"] * s["A2_OUT"] - s["A1_OUT"] ** 2

    return s



import numpy as np
from numba import njit


@njit(cache=True)
def compute_mad_quadrupole_matrix(element_parameters: list, global_parameters: list) -> np.ndarray:
    L: float = element_parameters[0]
    k1: float = element_parameters[1]
    R = np.zeros((6, 6))

    # Setup basic notations - Horizontal plane
    if k1 == 0:
        cx = 1
        sx = L
    elif k1 > 0:
        cx = np.cos(np.sqrt(k1) * L)
        sx = np.sin(np.sqrt(k1) * L) / np.sqrt(k1)
    else:
        cx = np.cosh(np.sqrt(-k1) * L)
        sx = np.sinh(np.sqrt(-k1) * L) / np.sqrt(-k1)

    # Setup basic notations - Vertical plane
    if k1 == 0:
        cy = 1
        sy = L
    elif k1 > 0:
        cy = np.cosh(np.sqrt(k1) * L)
        sy = np.sinh(np.sqrt(k1) * L) / np.sqrt(k1)
    else:
        cy = np.cos(np.sqrt(-k1) * L)
        sy = np.sin(np.sqrt(-k1) * L) / np.sqrt(-k1)

    # Definition of the matrix elements
    R[0, 0] = cx
    R[0, 1] = sx
    R[1, 0] = -(k1 * sx)
    R[1, 1] = cx
    R[2, 2] = cy
    R[2, 3] = sy
    R[3, 2] = k1 * sy
    R[3, 3] = cy
    R[4, 4] = 1
    R[5, 5] = 1

    return R


@njit(cache=True)
def compute_mad_quadrupole_tensor(element_parameters: list, global_parameters: list) -> np.ndarray:
    L: float = element_parameters[0]
    k1: float = element_parameters[1]
    beta: float = global_parameters[0]
    gamma = 1 / np.sqrt(1 - beta**2)
    T = np.zeros((6, 6, 6))

    # Setup basic notations - Horizontal plane
    if k1 == 0:
        cx = 1
        sx = L
    elif k1 > 0:
        cx = np.cos(np.sqrt(k1) * L)
        sx = np.sin(np.sqrt(k1) * L) / np.sqrt(k1)
    else:
        cx = np.cosh(np.sqrt(-k1) * L)
        sx = np.sinh(np.sqrt(-k1) * L) / np.sqrt(-k1)

    # Setup basic notations - Vertical plane
    if k1 == 0:
        cy = 1
        sy = L
    elif k1 > 0:
        cy = np.cosh(np.sqrt(k1) * L)
        sy = np.sinh(np.sqrt(k1) * L) / np.sqrt(k1)
    else:
        cy = np.cos(np.sqrt(-k1) * L)
        sy = np.sin(np.sqrt(-k1) * L) / np.sqrt(-k1)

    # Definition of the tensor elements
    T[0, 0, 5] = (k1 * L * sx) / (2 * beta)
    T[0, 1, 5] = (-(cx * L) - sx) / (2 * beta)
    T[1, 0, 5] = -(k1 * (-(cx * L) + sx)) / (2 * beta)
    T[1, 1, 5] = (k1 * L * sx) / (2 * beta)
    T[2, 2, 5] = -(k1 * L * sy) / (2 * beta)
    T[2, 3, 5] = (-(cy * L) - sy) / (2 * beta)
    T[3, 2, 5] = (k1 * (-(cy * L) + sy)) / (2 * beta)
    T[3, 3, 5] = -(k1 * L * sy) / (2 * beta)
    T[4, 0, 0] = -(k1 * (L - cx * sx)) / (4 * beta)
    T[4, 0, 1] = (k1 * sx**2) / (2 * beta)
    T[4, 1, 1] = -(L + cx * sx) / (4 * beta)
    T[4, 2, 2] = (k1 * (L - cy * sy)) / (4 * beta)
    T[4, 2, 3] = (k1 * sy**2) / (2 * beta)
    T[4, 3, 3] = -(L + cy * sy) / (4 * beta)
    T[4, 5, 5] = (-3 * L) / (2 * beta**3 * gamma**2)

    return T



"""
TODO
"""
from ... import ureg as _ureg
from .magnets import Drift as _Drift


class Collimator(_Drift):
    """
    Define a Collimator.

    Attributes:
        PARAMETERS (dict): Dictionary containing the parameters of the Collimator with their default values.

    Examples:
        >>> c1 = Collimator('C1', L=10*_ureg.cm, APERTYPE='ELIPTICAL', APERTURE=[2*_ureg.cm, 3*_ureg.cm])
        >>> c1 #doctest: +NORMALIZE_WHITESPACE
            Collimator: {'NAME': 'C1',
                         'AT_ENTRY': <Quantity(0, 'meter')>,
                         'AT_CENTER': <Quantity(0, 'meter')>,
                         'AT_EXIT': <Quantity(0, 'meter')>,
                         'L': <Quantity(10, 'centimeter')>,
                         'APERTYPE': 'ELIPTICAL',
                         'APERTURE': [<Quantity(2, 'centimeter')>, <Quantity(3, 'centimeter')>]}
    """

    PARAMETERS = {
        "APERTYPE": (None, "Aperture type (CIRCULAR, ELLIPTICAL, RECTANGULAR or PHASE_SPACE)"),
        "APERTURE": ([None, None, None, None], ""),
    }


class CircularCollimator(Collimator):
    """
    Define a CircularCollimator.

    Attributes:
        PARAMETERS (dict): Dictionary containing the parameters of the CircularCollimator with their default values.

    Examples:
        >>> c2 = CircularCollimator('C2', L=10*_ureg.cm, APERTURE=[2*_ureg.cm])
        >>> c2 #doctest: +NORMALIZE_WHITESPACE
            CircularCollimator: {'NAME': 'C2',
                         'AT_ENTRY': <Quantity(0, 'meter')>,
                         'AT_CENTER': <Quantity(0, 'meter')>,
                         'AT_EXIT': <Quantity(0, 'meter')>,
                         'L': <Quantity(10, 'centimeter')>,
                         'APERTYPE': 'CIRCULAR',
                         'APERTURE': [<Quantity(2, 'centimeter')>]}
    """

    PARAMETERS = {
        "APERTYPE": ("CIRCULAR", "Aperture type"),
    }


class EllipticalCollimator(Collimator):
    """
    Define an EllipticalCollimator.

    Attributes:
        PARAMETERS (dict): Dictionary containing the parameters of the EllipticalCollimator with their default values.

    Examples:
        >>> c2 = EllipticalCollimator('C2', L=10*_ureg.cm, APERTURE=[2*_ureg.cm, 3*_ureg.cm])
        >>> c2 #doctest: +NORMALIZE_WHITESPACE
            EllipticalCollimator: {'NAME': 'C2',
                         'AT_ENTRY': <Quantity(0, 'meter')>,
                         'AT_CENTER': <Quantity(0, 'meter')>,
                         'AT_EXIT': <Quantity(0, 'meter')>,
                         'L': <Quantity(10, 'centimeter')>,
                         'APERTYPE': 'ELLIPTICAL',
                         'APERTURE': [<Quantity(2, 'centimeter')>, <Quantity(3, 'centimeter')>]}
    """

    PARAMETERS = {
        "APERTYPE": ("ELLIPTICAL", "Aperture type"),
    }


class RectangularCollimator(Collimator):
    """
    Define a RectangularCollimator.

    Attributes:
        PARAMETERS (dict): Dictionary containing the parameters of the RectangularCollimator with their default values.

    Examples:
        >>> c3 = RectangularCollimator('C3', L=10*_ureg.cm, APERTURE=[2*_ureg.cm, 3*_ureg.cm])
        >>> c3 #doctest: +NORMALIZE_WHITESPACE
            RectangularCollimator: {'NAME': 'C3',
                         'AT_ENTRY': <Quantity(0, 'meter')>,
                         'AT_CENTER': <Quantity(0, 'meter')>,
                         'AT_EXIT': <Quantity(0, 'meter')>,
                         'L': <Quantity(10, 'centimeter')>,
                         'APERTYPE': 'RECTANGULAR',
                         'APERTURE': [<Quantity(2, 'centimeter')>, <Quantity(3, 'centimeter')>]}
    """

    PARAMETERS = {
        "APERTYPE": ("RECTANGULAR", "Aperture type"),
    }


class Dump(RectangularCollimator):
    """
    Define a Dump.

    Attributes:
        PARAMETERS (dict): Dictionary containing the parameters of the Dump with their default values.

    Examples:
        >>> d1 = Dump('D1', L=10*_ureg.cm)
        >>> d1 #doctest: +NORMALIZE_WHITESPACE
            Dump: {'NAME': 'D1',
                   'AT_ENTRY': <Quantity(0, 'meter')>,
                   'AT_CENTER': <Quantity(0, 'meter')>,
                   'AT_EXIT': <Quantity(0, 'meter')>,
                   'L': <Quantity(10, 'centimeter')>,
                   'APERTYPE': 'RECTANGULAR',
                   'APERTURE': <Quantity(0.0, 'centimeter')>}
    """

    PARAMETERS = {
        "APERTURE": [0.0 * _ureg.cm, 0.0 * _ureg.cm],
    }



"""
This code is from MAD-X twiss.f90. It is used to allow comparisons between the MAD-X tensor and
the generated code for MAD8 or Transport style tensor computation.
"""
import numpy as np
from numpy import cos, cosh, sin, sinh, sqrt

zero = 0
one = 1
two = 2
three = 3
four = 4
six = 6
nine = 9
twelve = 12
fifteen = 15
twty = 20
twty2 = 22
twty4 = 24
thty = 30
foty2 = 42
fvty6 = 56
svty2 = 72
httwty = 120
c1 = one
c2 = one / 2
c3 = one / twty4
c4 = one / 720
s1 = one
s2 = one / 6
s3 = one / httwty
s4 = one / 5040
cg0 = one / twty
cg1 = 5 / 840
cg2 = 21 / 60480
ch0 = one / fvty6
ch1 = 14 / 4032
ch2 = 147 / 443520


def tmfoc(el, sk1):
    #   !----------------------------------------------------------------------*
    #   ! Purpose:                                                             *
    #   !   Compute linear focussing functions.                                *
    #   ! Input:                                                               *
    #   !   el        (double)  element length.                                *
    #   !   sk1       (double)  quadrupole strength.                           *
    #   ! Output:                                                              *
    #   !   c         (double)  cosine-like function.             c(k,l)       *
    #   !   s         (double)  sine-like function.               s(k,l)       *
    #   !   d         (double)  dispersion function.              d(k,l)       *
    #   !   f         (double)  integral of dispersion function.  f(k,l)       *
    #   !----------------------------------------------------------------------*
    #   double precision, intent(IN)  :: el, sk1
    #   double precision, intent(OUT) :: c, s, d, f

    #   double precision :: qk, qkl, qkl2

    #   double precision, parameter :: twty=20d0, thty=30d0, foty2=42d0
    qk = sqrt(abs(sk1))
    qkl = qk * el
    qkl2 = sk1 * el**2
    if abs(qkl2) < 1e-2:
        c = one - qkl2 * (one - qkl2 / twelve) / two
        s = (one - qkl2 * (one - qkl2 / twty) / six) * el
        d = (one - qkl2 * (one - qkl2 / thty) / twelve) * el**2 / two
        f = (one - qkl2 * (one - qkl2 / foty2) / twty) * el**3 / six
    else:
        if qkl2 > zero:
            c = cos(qkl)
            s = sin(qkl) / qk
        else:
            c = cosh(qkl)
            s = sinh(qkl) / qk
        d = (one - c) / sk1
        f = (el - s) / sk1
    return c, s, d, f


def tmsect(fsec, el, h, sk1, sk2, dh, beta, gamma):
    # use twissbeamfi, only : beta, gamma, dtbyds
    # use matrices, only: EYE
    # use math_constfi, only :
    # implicit none
    # !----------------------------------------------------------------------*
    # !     Purpose:                                                         *
    # !     TRANSPORT map for a sector dipole without fringe fields.         *
    # !     Input:                                                           *
    # !     fsec      (logical) if true, return second order terms.          *
    # !     el        (double)  element length.                              *
    # !     h         (double)  reference curvature of magnet.               *
    # !     dh        (double)  dipole field error.                          *
    # !     sk1       (double)  quadrupole strength.                         *
    # !     sk2       (double)  sextupole strengh.                           *
    # !     Output:                                                          *
    # !     ek(6)     (double)  kick due to dipole.                          *
    # !     re(6,6)   (double)  transfer matrix.                             *
    # !     te(6,6,6) (double)  second order terms.                          *
    # !----------------------------------------------------------------------*
    # logical, intent(IN) :: fsec
    # double precision :: el, h, dh, sk1, sk2
    # double precision :: ek(6), re(6,6), te(6,6,6)

    # double precision :: bi, bi2, bi2gi2
    # double precision :: cm, cp, cx, cy, cyy, dd, difsq, dm, dp, dx, dyy
    # double precision :: fm, fp, fx, fyy, gx, h2, hx, sm, sp, sumsq, sx, sy, syy
    # double precision :: t1, t116, t126, t166, t2, t216, t226, t266
    # double precision :: t336, t346, t436, t446, t5, t516, t526, t566
    # double precision :: xk, xkl, xklsq, xksq, xs6
    # double precision :: y0, y1, y2, y2klsq, y2ksq, yk, ykl, yklsq, yksq, ys2
    # double precision :: zc, zd, zf, zs

    # !---- Initialize.
    ek = np.zeros(6)
    re = np.eye(6)
    te = np.zeros((6, 6, 6))

    bi = 1 / beta
    bi2 = bi * bi
    bi2gi2 = one / (beta * gamma) ** 2

    # !---- Horizontal.
    xksq = h**2 + sk1
    xk = sqrt(abs(xksq))
    xkl = xk * el
    xklsq = xksq * el**2

    if abs(xklsq) < 1e-2:
        cx = c1 - xklsq * (c2 - xklsq * c3)
        sx = (s1 - xklsq * (s2 - xklsq * s3)) * el
        dx = (c2 - xklsq * (c3 - xklsq * c4)) * el**2
        fx = (s2 - xklsq * (s3 - xklsq * s4)) * el**3
        gx = (cg0 - xklsq * (cg1 - xklsq * cg2)) * el**5
        hx = (ch0 - xklsq * (ch1 - xklsq * ch2)) * el**7
    else:
        if xklsq > 0:
            cx = cos(xkl)
            sx = sin(xkl) / xk
        else:
            cx = cosh(xkl)
            sx = sinh(xkl) / xk

        dx = (one - cx) / xksq
        fx = (el - sx) / xksq
        gx = (three * el - sx * (four - cx)) / (two * xksq**2)
        hx = (fifteen * el - sx * (twty2 - nine * cx + two * cx**2)) / (six * xksq**3)

    re[0, 0] = cx
    re[0, 1] = sx
    re[0, 5] = h * dx * bi
    re[1, 0] = -xksq * sx
    re[1, 1] = cx
    re[1, 5] = h * sx * bi
    re[4, 1] = -re[0, 5]
    re[4, 0] = -re[1, 5]
    re[4, 5] = el * bi2gi2 - h**2 * fx * bi2

    ek[0] = -dh * dx
    ek[1] = -dh * sx
    ek[4] = h * dh * fx * bi + el  # *dtbyds

    # !---- Vertical.
    yksq = -sk1
    yk = sqrt(abs(yksq))
    ykl = yk * el
    yklsq = yksq * el**2

    if abs(yklsq) < 1e-2:
        cy = c1 - yklsq * (c2 - yklsq * c3)
        sy = (s1 - yklsq * (s2 - yklsq * s3)) * el
    elif yklsq > 0:
        cy = cos(ykl)
        sy = sin(ykl) / yk
    else:
        cy = cosh(ykl)
        sy = sinh(ykl) / yk

    re[2, 2] = cy
    re[2, 3] = sy
    re[3, 2] = -yksq * sy
    re[3, 3] = cy

    ek[2] = 0
    ek[3] = 0

    # !---- Second-order terms.
    if fsec:
        # !---- Pure horizontal terms.
        xs6 = (sk2 + two * h * sk1) / six
        ys2 = (sk2 + h * sk1) / two
        h2 = h / two

        t116 = xs6 * (three * sx * fx - dx**2) - h * sx**2
        t126 = xs6 * (sx * dx**2 - two * cx * gx) - h * sx * dx
        t166 = xs6 * (dx**3 - two * sx * gx) - h2 * dx**2
        t216 = xs6 * (three * cx * fx + sx * dx)
        t226 = xs6 * (three * sx * fx + dx**2)
        t266 = xs6 * (sx * dx**2 - two * cx * gx)
        t516 = h * xs6 * (three * dx * fx - four * gx) + (sk1 / two) * (fx + sx * dx)
        t526 = h * xs6 * (dx**3 - two * sx * gx) + (sk1 / two) * dx**2
        t566 = h * xs6 * (three * hx - two * dx * gx) + (sk1 / two) * gx - fx

        t1 = (sk1 / two) * (dx**2 - sx * fx) - dx
        t2 = (sk1 / two) * (el * dx - fx)
        t5 = fx - sk1 * (gx - fx * dx / two)

        te[0, 0, 0] = -xs6 * (sx**2 + dx) - h2 * xksq * sx**2
        te[0, 0, 1] = (-xs6 * dx + h2 * cx) * sx
        te[0, 1, 1] = (-xs6 * dx + h2 * cx) * dx
        te[0, 0, 5] = (-h2 * t116 + (sk1 / four) * el * sx) * bi
        te[0, 1, 5] = (-h2 * t126 + (sk1 / four) * (el * dx - fx) - sx / two) * bi
        te[0, 5, 5] = (-(h**2) * t166 + h * t1) * bi2 - h2 * dx * bi2gi2
        te[1, 0, 0] = -xs6 * (one + two * cx) * sx
        te[1, 0, 1] = -xs6 * (one + two * cx) * dx
        te[1, 1, 1] = -(two * xs6 * dx + h2) * sx
        te[1, 0, 5] = (-h2 * t216 - (sk1 / four) * (sx - el * cx)) * bi
        te[1, 1, 5] = (-h2 * t226 + (sk1 / four) * el * sx) * bi
        te[1, 5, 5] = (-(h**2) * t266 + h * t2) * bi2 - h2 * sx * bi2gi2

        te[4, 0, 0] = (h2 * xs6 * (sx * dx + three * fx) - (sk1 / four) * (el - cx * sx)) * bi
        te[4, 0, 1] = (h2 * xs6 * dx**2 + (sk1 / four) * sx**2) * bi
        te[4, 1, 1] = (h * xs6 * gx - sk1 * (fx - sx * dx) / four - sx / two) * bi
        te[4, 0, 5] = h2 * ((t516 - sk1 * (el * dx - fx) / two) * bi2 + sx * bi2gi2)
        te[4, 1, 5] = h2 * ((t526 - sk1 * (dx**2 - sx * fx) / two) * bi2 + dx * bi2gi2)
        te[4, 5, 5] = (h**2 * (t566 + t5) * bi2 + (three / two) * (h**2 * fx - el) * bi2gi2) * bi

        # !---- Mixed terms.
        y2ksq = four * yksq
        cyy, syy, dyy, fyy = tmfoc(el, y2ksq)
        y2klsq = y2ksq * el**2
        if max(abs(y2klsq), abs(xklsq)) < 1e-2:
            y0 = one
            y1 = xklsq + y2klsq
            y2 = xklsq**2 + xklsq * y2klsq + y2klsq**2
            zc = (y0 - (y1 - y2 / thty) / twelve) * el**2 / two
            zs = (y0 - (y1 - y2 / foty2) / twty) * el**3 / six
            zd = (y0 - (y1 - y2 / fvty6) / thty) * el**4 / twty4
            zf = (y0 - (y1 - y2 / svty2) / foty2) * el**5 / httwty
        elif xksq <= zero or yksq <= zero:
            dd = xksq - y2ksq
            zc = (cyy - cx) / dd
            zs = (syy - sx) / dd
            zd = (dyy - dx) / dd
            zf = (fyy - fx) / dd
        else:
            sumsq = (xk / two + yk) ** 2
            difsq = (xk / two - yk) ** 2
            cp, sp, dp, fp = tmfoc(el, sumsq)
            cm, sm, dm, fm = tmfoc(el, difsq)
            zc = sp * sm / two
            zs = (sp * cm - cp * sm) / (four * xk * yk)
            if xksq > y2ksq:
                zd = (dyy - zc) / xksq
                zf = (fyy - zs) / xksq
            else:
                zd = (dx - zc) / y2ksq
                zf = (fx - zs) / y2ksq

        t336 = sk2 * (cy * zd - two * sk1 * sy * zf) + h * sk1 * fx * sy
        t346 = sk2 * (sy * zd - two * cy * zf) + h * fx * cy
        t436 = two * ys2 * fx * cy - sk2 * sk1 * (sy * zd - two * cy * zf)
        t446 = two * ys2 * fx * sy - sk2 * (cy * zd - two * sk1 * sy * zf)

        te[0, 2, 2] = +sk2 * sk1 * zd + ys2 * dx
        te[0, 2, 3] = +sk2 * zs / two
        te[0, 3, 3] = +sk2 * zd - h2 * dx
        te[1, 2, 2] = +sk2 * sk1 * zs + ys2 * sx
        te[1, 2, 3] = +sk2 * zc / two
        te[1, 3, 3] = +sk2 * zs - h2 * sx
        te[2, 0, 2] = +sk2 * (cy * zc / two - sk1 * sy * zs) + h2 * sk1 * sx * sy
        te[2, 0, 3] = +sk2 * (sy * zc / two - cy * zs) + h2 * sx * cy
        te[2, 1, 2] = +sk2 * (cy * zs / two - sk1 * sy * zd) + h2 * sk1 * dx * sy
        te[2, 1, 3] = +sk2 * (sy * zs / two - cy * zd) + h2 * dx * cy
        te[2, 2, 5] = (h2 * t336 - sk1 * el * sy / four) * bi
        te[2, 3, 5] = (h2 * t346 - (sy + el * cy) / four) * bi
        te[3, 0, 2] = sk2 * sk1 * (cy * zs - sy * zc / two) + ys2 * sx * cy
        te[3, 0, 3] = sk2 * (sk1 * sy * zs - cy * zc / two) + ys2 * sx * sy
        te[3, 1, 2] = sk2 * sk1 * (cy * zd - sy * zs / two) + ys2 * dx * cy
        te[3, 1, 3] = sk2 * (sk1 * sy * zd - cy * zs / two) + ys2 * dx * sy
        te[3, 2, 5] = (h2 * t436 + sk1 * (sy - el * cy) / four) * bi
        te[3, 3, 5] = (h2 * t446 - sk1 * el * sy / four) * bi

        te[4, 2, 2] = (-h * sk2 * sk1 * zf - h * ys2 * fx + sk1 * (el - cy * sy) / four) * bi
        te[4, 2, 3] = (-h * sk2 * zd / two - sk1 * sy**2 / four) * bi
        te[4, 3, 3] = (-h * sk2 * zf + h * h2 * fx - (el + sy * cy) / four) * bi
        # call tmsymm(te)

    return re, te



import os
import sys

import georges_core.kinematics as gkin
import numpy as np
import pandas as pd
import pybdsim
from compute_quantiles import compute_quantile_2dim, compute_quantile_4dim
from lmfit import Model, Parameters

from georges import ureg as _ureg


def cut_data(filepath: str = None, epos: float = 100, matname: str = None, nprimary: int = 10) -> pd.DataFrame:
    """

    Args:
        filepath (str): Path to the file
        epos (float): Theoritical energy at the exit of the degrader
        matname (str): Name of the material
        nprimary (int): Number of particles launched in BDSIM.

    Returns:

    """

    element = "DEG"
    coordinates = ["x", "y", "xp", "yp", "energy", "p", "partID", "parentID", "S"]

    data = pybdsim.Analysis.BDSimOutput(filepath).event.samplers[element].df
    data = data[coordinates]

    data.reset_index(drop=True, inplace=True)

    data.query("parentID == 0 and partID == 2212", inplace=True)  # only get the primaries
    data["energy"] = gkin.etot_to_ekin(data["energy"].values * _ureg.GeV).m_as("MeV")
    data["momentum"] = data["p"] * 1000  # In Mev/c
    data["xp"] *= 1000
    data["yp"] *= 1000
    data["x"] *= 1000
    data["y"] *= 1000

    # Cut the data
    if epos < 220:
        quantile, data_cutted = compute_quantile_4dim(data)

    else:
        quantile, data_cutted = compute_quantile_2dim(data)

    # Properties of the cutted data
    momentum_cut = data_cutted["momentum"].mean()
    deviation_cut = (data_cutted["momentum"] - momentum_cut) / momentum_cut
    dpp_mean_cut = np.mean(deviation_cut)
    dpp_rms_cut = np.std(deviation_cut)
    transmission_cut = len(data_cutted) / nprimary

    # Return the results
    return pd.DataFrame(
        data={
            "MatName": [matname],
            "Epost": [epos],
            "DPPmean_cut": [dpp_mean_cut],
            "DPPrms_cut": [dpp_rms_cut],
            "Transmission_cut": [transmission_cut],
        },
    )


def fit_values(x, **params):
    val = 0.0
    parnames = sorted(params.keys())
    for i, pname in enumerate(parnames):
        val += params[pname] * x**i
    return val


if __name__ == "__main__":
    path = sys.argv[1]
    nparticles = int(sys.argv[2])

    results = pd.DataFrame()
    for file in os.listdir(path):
        if "parquet" in file:
            print(f"Process {file}")
            results = pd.concat(
                [
                    results,
                    cut_data(
                        filepath=os.path.join(path, file),
                        epos=float(file.split("-")[2].replace("E", "").replace("root", "")),
                        matname=file.split("-")[1],
                        nprimary=nparticles,
                    ),
                ],
                axis=0,
            )

    params_transmission = Parameters()
    params_transmission.add("C0", value=1)
    params_transmission.add("C1", value=1)
    params_transmission.add("C2", value=1)
    params_transmission.add("C3", value=0)

    fit_data = Model(fit_values)
    data_transmission = results[["Epost", "Transmission_cut"]]
    fit_transmission = fit_data.fit(
        data_transmission["Transmission_cut"],
        params_transmission,
        x=data_transmission["Epost"],
    )
    coeff_transmission = fit_transmission.best_values

    params_dpp = Parameters()
    params_dpp.add("C0", value=1)
    params_dpp.add("C1", value=1)
    params_dpp.add("C2", value=1)
    params_dpp.add("C3", value=0)
    params_dpp.add("C4", value=0)
    params_dpp.add("C5", value=0)
    data_dpp = results[["Epost", "DPPrms_cut"]]
    fit_dpp = fit_data.fit(data_dpp["DPPrms_cut"], params_dpp, x=data_dpp["Epost"])
    coeff_dpp = fit_dpp.best_values

    print(
        f"""
    materiaName,
    {coeff_transmission['C0']},{coeff_transmission['C1']},{coeff_transmission['C2']},{coeff_transmission['C3']}
    {coeff_dpp['C0']},{coeff_dpp['C1']},{coeff_dpp['C2']},{coeff_dpp['C3']},{coeff_dpp['C4']},{coeff_dpp['C5']}
    """,
    )



from importlib.metadata import version

__version__ = version("georges")

from georges_core import Q_, Kinematics, particles, ureg
from georges_core.distribution import *
from georges_core.madx import madx
from georges_core.sequences import *
from georges_core.sequences.betablock import BetaBlock

from .vis import ManzoniMatplotlibArtist, ManzoniPlotlyArtist



"""
The beam propagation from one element to another in the beamline is implemented in the file `core.py`.
The `track` function contains the loop over the different elements, with optionally the check of
the apertures to select the particles that survive the tracking at the end of each element,
and the use of “observers” if defined by the user, to save the data during the tracking.
The file `core.py` also contains a Twiss function, which allows the calculation of the matrix elements
of all the elements along a given beamline for the Twiss functions calculation based on the 11 particles
method. The user must be aware that this function also needs the georges_core module to work properly,
as the Twiss computation is done in the end in the georges_core library, using the matrix elements
calculated in georges.
    """

from __future__ import annotations

from typing import TYPE_CHECKING, List, Optional

import numpy as _np
import pandas as _pd
from georges_core.sequences import BetaBlock as _BetaBlock
from georges_core.twiss import Twiss as _Twiss
from numba.typed import List as nList

from .beam import Beam as _Beam
from .observers import BeamObserver as _BeamObserver

if TYPE_CHECKING:
    from .. import Kinematics as _Kinematics
    from .input import Input as _Input
    from .observers import Observer as _Observer


def track(
    beamline: _Input,
    beam: _Beam,
    observers: List[Optional[_Observer]] = None,
    check_apertures_exit: bool = False,
    check_apertures_entry: bool = False,
):
    """
    Args:
        beamline:
        beam:
        observers:
        check_apertures_exit:
        check_apertures_entry:
    Returns:
    """
    if observers is None:
        observers = []
    global_parameters = nList()
    global_parameters.append(beam.kinematics.beta)
    b1 = _np.copy(beam.distribution)
    b2 = _np.zeros(b1.shape)
    for e in beamline.sequence:
        if check_apertures_entry:
            b2, b1 = e.check_aperture(b2, b1)
            if b1.shape != b2.shape:
                b1 = _np.zeros(b2.shape)
            if b2.shape[0] == 0:
                break
        b1, b2 = e.propagate(b1, b2, global_parameters)
        if check_apertures_exit:
            b1, b2 = e.check_aperture(b1, b2)
        for o in observers:
            if o is not None:
                o(e, b1, b2)
        if b1.shape != b2.shape:
            b1 = _np.zeros(b2.shape)
        if b2.shape[0] == 0:
            break
        b2, b1 = b1, b2


def twiss(
    beamline: _Input,
    kinematics: _Kinematics,
    reference_particle: _np.ndarray = None,
    offsets=None,
    twiss_parametrization: bool = True,
    twiss_init: _BetaBlock = None,
    with_phase_unrolling: bool = True,
) -> _pd.DataFrame:
    """

    Args:
        beamline:
        kinematics:
        reference_particle:
        offsets:
        twiss_parametrization:
        twiss_init:
        with_phase_unrolling:
    Returns:

    """

    def track_for_twiss() -> _pd.DataFrame:
        nonlocal reference_particle
        nonlocal offsets

        if reference_particle is None:
            reference_particle = _np.array([0.0, 0.0, 0.0, 0.0, 0.0, 0.0])
        if offsets is None:
            offsets = _np.array([0.01, 0.01, 0.01, 0.01, 0.01])
        pt = _Beam.compute_pt(dpp=offsets[4], beta=kinematics.beta)
        coordinates = _np.array(
            [
                reference_particle,
                reference_particle + [offsets[0], 0.0, 0.0, 0.0, 0.0, 0.0],
                reference_particle + [0.0, offsets[1], 0.0, 0.0, 0.0, 0.0],
                reference_particle + [0.0, 0.0, offsets[2], 0.0, 0.0, 0.0],
                reference_particle + [0.0, 0.0, 0.0, offsets[3], 0.0, 0.0],
                reference_particle + [0.0, 0.0, 0.0, 0.0, offsets[4], pt],
                reference_particle + [-offsets[0], 0.0, 0.0, 0.0, 0.0, 0.0],
                reference_particle + [0.0, -offsets[1], 0.0, 0.0, 0.0, 0.0],
                reference_particle + [0.0, 0.0, -offsets[2], 0.0, 0.0, 0.0],
                reference_particle + [0.0, 0.0, 0.0, -offsets[3], 0.0, 0.0],
                reference_particle + [0.0, 0.0, 0.0, 0.0, -offsets[4], -pt],
            ],
        )
        beam = _Beam(kinematics=kinematics, distribution=coordinates)
        observer = _BeamObserver(with_input_beams=False)
        track(beam=beam, beamline=beamline, observers=[observer])
        return observer.to_df()

    def compute_matrix_for_twiss(data: _pd.DataFrame) -> _pd.DataFrame:
        # Make some changes in the format of the matrices to be consistent with georges-core
        # Indeed in georges core, we have a longitudinal coordinates
        normalization = 2 * offsets
        normalization = _np.insert(normalization, -1, normalization[0], 0)
        _matrix = {}
        for label, d in data.iterrows():
            m = d["BEAM_OUT"]
            m[:, 4] = m[:, 5]
            m = m[:, 0:5]
            m = _np.hstack((m, _np.zeros((m.shape[0], 1))))
            m[:, [-2, -1]] = m[:, [-1, -2]]
            m = _np.insert(m, 5, _np.zeros(m.shape[1]), 0)
            m = _np.insert(m, -1, _np.zeros(m.shape[1]), 0)
            _matrix[label] = {
                f"R{j + 1}{i + 1}": (m[i + 1, j] - m[i + 1 + 6, j]) / normalization[i]
                for j in range(0, 6)
                for i in range(0, 6)
            }
            _matrix[label]["S"] = d["AT_EXIT"].m_as("m")
        return _pd.DataFrame.from_dict(_matrix, orient="index")

    tracks = track_for_twiss()
    matrix = compute_matrix_for_twiss(tracks)
    if twiss_parametrization:
        matrix = _Twiss(twiss_init=twiss_init, with_phase_unrolling=with_phase_unrolling)(matrix=matrix)
    return matrix


def match(beamline: _Input, beam: _Beam):
    ...



"""
The file `Beam.py` contains the implementation of the class beam for the tracking with Manzoni.
A beam definition requires distribution and kinematics to allow the generation of the initial
particles for tracking.
"""
import numpy as np
from georges_core import Kinematics as _Kinematics


class Beam:
    def __init__(self, kinematics: _Kinematics, distribution: np.ndarray):
        self._kinematics = kinematics
        self._distribution = distribution.astype(float, copy=False)

    @classmethod
    def compute_pt(cls, dpp: np.ndarray, beta: float, first_order: bool = False) -> np.ndarray:
        """

        Args:
            dpp:
            beta: the relativistic beta
            first_order: approximate p_t from dpp at first order (valid only for beta close to 1)

        Returns:

        """
        if first_order:
            return dpp * beta
        else:
            return (-(2 / beta) + np.sqrt((2 / beta) ** 2 + 4 * (dpp**2 + 2 * dpp))) / 2

    @classmethod
    def compute_dpp(cls, pt: np.ndarray, beta: float, first_order: bool = False) -> np.ndarray:
        """

        Args:
            pt:
            beta: the relativistic beta
            first_order: approximate dpp from p_t at first order (valid only for beta close to 1)

        Returns:

        """
        if first_order:
            return pt / beta
        else:
            return np.sqrt(pt**2 + 2 * pt / beta + 1) - 1

    @property
    def kinematics(self) -> _Kinematics:
        return self._kinematics

    @property
    def distribution(self) -> np.ndarray:
        return self._distribution

    @distribution.setter
    def distribution(self, dist: np.ndarray):
        self._distribution = dist


class MadXBeam(Beam):
    def __init__(self, kinematics: _Kinematics, distribution: np.ndarray, first_order: bool = False):
        super().__init__(kinematics=kinematics, distribution=distribution)
        pt = Beam.compute_pt(
            distribution[:, -1],
            kinematics.beta,
            first_order=first_order,
        )
        self._distribution = np.zeros((len(distribution), 6))
        self._distribution[:, :-1] = distribution
        self._distribution[:, -1] = pt


class TransportBeam(Beam):
    def __init__(self, kinematics: _Kinematics, distribution: np.ndarray):
        super().__init__(kinematics=kinematics, distribution=distribution)



import numpy as np
from numba import njit
from numba.typed import List as nList
from numpy import cos, sin, tan


@njit(cache=True)
def compute_transport_fringe_in_matrix(element_parameters: nList) -> np.ndarray:
    L: float = element_parameters[0]
    alpha: float = element_parameters[1]
    beta1: float = element_parameters[3]
    gap: float = element_parameters[4]
    fint: float = element_parameters[5]
    h = alpha / L
    R = np.zeros((6, 6))
    phi1 = fint * h * gap * cos(beta1) ** (-1) * (1 + sin(beta1) ** 2)
    R[0, 0] = 1
    R[1, 0] = h * tan(beta1)
    R[1, 1] = 1
    R[2, 2] = 1
    R[3, 2] = -h * tan(beta1 - phi1)
    R[3, 3] = 1
    R[4, 4] = 1
    R[5, 5] = 1
    return R


@njit(cache=True)
def compute_transport_fringe_in_tensor(element_parameters: nList) -> np.ndarray:
    L: float = element_parameters[0]
    alpha: float = element_parameters[1]
    k1: float = element_parameters[2]
    beta1: float = element_parameters[3]
    gap: float = element_parameters[4]
    fint: float = element_parameters[5]
    r1: float = element_parameters[6]
    h = alpha / L

    T = np.zeros((6, 6, 6))
    phi1 = fint * h * gap * cos(beta1) ** (-1) * (1 + sin(beta1) ** 2)
    T[0, 0, 0] = -h * tan(beta1) ** 2 / 2
    T[0, 2, 2] = h * cos(beta1) ** (-1) ** 2 / 2
    T[1, 0, 0] = k1 * tan(beta1) + h * cos(beta1) ** (-1) ** 3 / (2 * r1)
    T[1, 0, 1] = h * tan(beta1) ** 2
    T[1, 0, 4] = -h * tan(beta1)
    T[1, 2, 2] = (
        -k1 * tan(beta1) + 0.5 * h**2 * tan(beta1) + h**2 + tan(beta1) ** 3 - h * cos(beta1) ** (-1) ** 3 / (2 * r1)
    )
    T[1, 2, 3] = -h * tan(beta1) ** 2
    T[2, 0, 2] = h * tan(beta1) ** 2
    T[3, 0, 2] = -2 * k1 * tan(beta1) - h * cos(beta1) ** (-1) ** 3 / r1
    T[3, 0, 3] = -h * tan(beta1) ** 2
    T[3, 1, 2] = -h * cos(beta1) ** (-1) ** 2
    T[3, 2, 4] = -h * phi1 * cos(beta1 - phi1) ** (-2) + h * tan(beta1)
    return T



from .materials import __initialize_materials_database
from .mcs import ICRU, DifferentialHighland, DifferentialMoliere, FermiRossi, ICRUProtons
from .propagation import propagate, track_energy

__initialize_materials_database()



"""
TODO
"""
from typing import List, Tuple

import numpy as _np

try:
    import numpy.random_intel as nprandom
except ModuleNotFoundError:
    import numpy.random as nprandom

from ... import Kinematics
from ... import ureg as _ureg
from ...fermi import materials
from ..beam import Beam as _Beam
from ..integrators import MadXIntegrator, MadXParaxialDriftIntegrator
from .elements import ManzoniElement as _ManzoniElement


class MaterialElement(_ManzoniElement):
    INTEGRATOR = None

    @property
    def material(self):
        if isinstance(self.MATERIAL, str):
            return getattr(materials, self.MATERIAL.capitalize())
        return self.MATERIAL

    @property
    def degraded_energy(self):
        if self.L.magnitude == 0.0:
            return self.KINETIC_ENERGY
        else:
            return (self.material.stopping(self.L, self.KINETIC_ENERGY)).ekin

    @property
    def cache(self) -> list:
        if not self.frozen:
            self._cache = self.parameters
        return self._cache

    @property
    def beta(self):
        return Kinematics(self.KINETIC_ENERGY, kinetic=True).beta

    @staticmethod
    def pseudo_aperture_check(beam, beta):
        px = beam[:, 1]
        py = beam[:, 3]
        pt = beam[:, 5]
        condition = 1 + 2 * pt / beta + pt**2 - px**2 - py**2 >= 0
        checked_beam = _np.compress(condition, beam, axis=0)

        return checked_beam


class Scatterer(MaterialElement):
    """
    Define a Scatterer.

    Attributes:
        PARAMETERS (dict): Dictionary containing the parameters of the Scatterer with their default values.

    Examples:
        >>> s1 = Scatterer('S1', MATERIAL=materials.Beryllium, KINETIC_ENERGY=230*_ureg.MeV)
        >>> s1 #doctest: +NORMALIZE_WHITESPACE
            Scatterer: {'NAME': 'S1',
                        'AT_ENTRY': <Quantity(0, 'meter')>,
                        'AT_CENTER': <Quantity(0, 'meter')>,
                        'AT_EXIT': <Quantity(0, 'meter')>,
                        'MATERIAL': <class 'georges.fermi.materials.Beryllium'>,
                        'KINETIC_ENERGY': <Quantity(230, 'megaelectronvolt')>,
                        'L': <Quantity(0.0, 'meter')>}
    """

    PARAMETERS = {
        "MATERIAL": (materials.Vacuum, "Degrader material"),
        "KINETIC_ENERGY": (0.0 * _ureg.MeV, "Incoming beam energy"),
        "L": (0.0 * _ureg.m, "Degrader length"),
    }

    @property
    def parameters(self) -> List[float]:
        fe = self.material.scattering(
            kinetic_energy=self.KINETIC_ENERGY,
            thickness=self.L,
            compute_a1=False,
            compute_a2=False,
        )
        return [
            fe["A"][0],
        ]

    def propagate(
        self,
        beam_in: _np.ndarray,
        beam_out: _np.ndarray = None,
        global_parameters: list = None,
    ) -> Tuple[_np.ndarray, _np.ndarray]:
        if self.material is materials.Vacuum:
            _np.copyto(dst=beam_out, src=beam_in, casting="no")
            return beam_in, beam_out
        a0 = self.cache[0]

        beam_out[:, 0] = beam_in[:, 0]
        beam_out[:, 2] = beam_in[:, 2]
        beam_out[:, 4] = beam_in[:, 4]
        beam_out[:, 5] = beam_in[:, 5]

        beam_out[:, 1] = beam_in[:, 1] + nprandom.normal(0.0, a0, size=beam_in.shape[0])
        beam_out[:, 3] = beam_in[:, 3] + nprandom.normal(0.0, a0, size=beam_in.shape[0])

        beam_out = self.pseudo_aperture_check(beam_out, self.beta)

        return beam_in, beam_out


class Degrader(MaterialElement):
    """
    Define a Degrader.

    Attributes:
        PARAMETERS (dict): Dictionary containing the parameters of the Degrader with their default values.

    Examples:
        >>> d1 = Degrader('D1', MATERIAL=materials.Beryllium, L=5*_ureg.cm, KINETIC_ENERGY=230*_ureg.MeV,
        ...               WITH_LOSSES=True)
        >>> d1 #doctest: +NORMALIZE_WHITESPACE
            Degrader: {'NAME': 'D1',
                       'AT_ENTRY': <Quantity(0, 'meter')>,
                       'AT_CENTER': <Quantity(0, 'meter')>,
                       'AT_EXIT': <Quantity(0, 'meter')>,
                       'MATERIAL': <class 'georges.fermi.materials.Beryllium'>,
                       'KINETIC_ENERGY': <Quantity(230, 'megaelectronvolt')>,
                       'L': <Quantity(5, 'centimeter')>,
                       'WITH_LOSSES': True}
    """

    PARAMETERS = {
        "MATERIAL": (materials.Vacuum, "Degrader material"),
        "KINETIC_ENERGY": (0.0 * _ureg.MeV, "Incoming beam energy"),
        "L": (0.0 * _ureg.m, "Degrader length"),
        "WITH_LOSSES": (False, "Boolean to compute losses and dpp"),
    }
    """Parameters of the element, with their default value and their descriptions."""

    @property
    def parameters(self) -> List[float]:
        fe = self.material.scattering(kinetic_energy=self.KINETIC_ENERGY, thickness=self.L)
        return [
            self.L.m_as("m"),
            fe["A"][0],
            fe["A"][1],
            fe["A"][2],
            self.material.energy_dispersion(energy=self.degraded_energy),
            self.material.losses(energy=self.degraded_energy),
        ]

    def propagate(
        self,
        beam_in: _np.ndarray,
        beam_out: _np.ndarray = None,
        global_parameters: list = None,
    ) -> Tuple[_np.ndarray, _np.ndarray]:
        length, a0, a1, a2, dpp, losses = self.cache

        if length == 0:
            _np.copyto(dst=beam_out, src=beam_in, casting="no")
            return beam_in, beam_out

        # Monte-Carlo method
        # Remove particles
        if self.WITH_LOSSES is True and losses != 1:
            idx = _np.random.randint(beam_in.shape[0], size=int(losses * beam_in.shape[0]))
            beam_out = beam_in[idx, :]
        else:
            beam_out = beam_in

        # Transport matrix
        matrix = _np.array(
            [
                [1, length, 0, 0, 0, 0],
                [0, 1, 0, 0, 0, 0],
                [0, 0, 1, length, 0, 0],
                [0, 0, 0, 1, 0, 0],
                [0, 0, 0, 0, 1, 0],
                [0, 0, 0, 0, 0, 1],
            ],
        )
        beam_out = beam_out.dot(matrix.T)

        # If the integrator is not MAD-X based, swap the last 2 columns
        if self.integrator not in [MadXIntegrator, MadXParaxialDriftIntegrator]:
            beam_out[:, [-2, -1]] = beam_out[:, [-1, -2]]

        # Interactions
        if self.material is not materials.Vacuum:
            beam_out += nprandom.multivariate_normal(
                [0.0, 0.0, 0.0, 0.0, 0.0, 0.0],
                _np.array(
                    [
                        [a2, a1, 0.0, 0.0, 0.0, 0.0],
                        [a1, a0, 0.0, 0.0, 0.0, 0.0],
                        [0.0, 0.0, a2, a1, 0.0, 0.0],
                        [0.0, 0.0, a1, a0, 0.0, 0.0],
                        [0.0, 0.0, 0.0, 0.0, dpp**2, 0.0],
                        [0.0, 0.0, 0.0, 0.0, 0.0, 0.0],
                    ],
                ),
                int(beam_out.shape[0]),
            )

        if self.integrator not in [MadXIntegrator, MadXParaxialDriftIntegrator]:
            beam_out[:, [-1, -2]] = beam_out[:, [-2, -1]]
        else:
            beam_out[:, -1] = _Beam.compute_pt(beam_out[:, -2], self.beta, first_order=False)
            beam_out = self.pseudo_aperture_check(beam_out, self.beta)

        return beam_in, beam_out


class BeamStop(MaterialElement):
    """
    Define a BeamStop.

    Attributes:
        PARAMETERS (dict): Dictionary containing the parameters of the BeamStop with their default values.

    Examples:
        >>> b1 = BeamStop('B1', MATERIAL=materials.Beryllium, L=5*_ureg.cm, KINETIC_ENERGY=230*_ureg.MeV,
        ...               RADIUS=3*_ureg.cm)
        >>> b1 #doctest: +NORMALIZE_WHITESPACE
            BeamStop: {'NAME': 'B1',
                       'AT_ENTRY': <Quantity(0, 'meter')>,
                       'AT_CENTER': <Quantity(0, 'meter')>,
                       'AT_EXIT': <Quantity(0, 'meter')>,
                       'MATERIAL': <class 'georges.fermi.materials.Beryllium'>,
                       'L': <Quantity(5, 'centimeter')>,
                       'RADIUS': <Quantity(3, 'centimeter')>,
                       'KINETIC_ENERGY': <Quantity(230, 'megaelectronvolt')>}
    """

    PARAMETERS = {
        "MATERIAL": (materials.Vacuum, "Beam Stop material"),
        "L": (0.0 * _ureg.m, "Beam Stop length"),
        "RADIUS": (0.0 * _ureg.m, "Beam Stop radius"),
        "KINETIC_ENERGY": (0.0 * _ureg.MeV, "Incoming beam energy"),
    }

    @property
    def parameters(self) -> list:
        return [
            self.L.m_as("m"),
            self.RADIUS.m_as("m"),
        ]

    def propagate(
        self,
        beam_in: _np.ndarray,
        beam_out: _np.ndarray = None,
        global_parameters: list = None,
    ) -> Tuple[_np.ndarray, _np.ndarray]:
        length, radius = self.parameters

        if length == 0 or radius == 0:
            _np.copyto(dst=beam_out, src=beam_in, casting="no")
            return beam_in, beam_out

        else:
            beam_out = _np.compress(
                (beam_in[:, 0] ** 2 + beam_in[:, 2] ** 2) > radius**2,
                beam_in,
                axis=0,
            )
            return beam_in, beam_out


class Gap(Degrader):
    pass



"""
The file `kernels.py` contains the loops that are the core of the particles propagation based on
their coordinates. Different batches are available, to allow a matrix (order 1) propagation,
a tensor (order 2) propagation or a matrix followed by a tensor (orders 1+2) propagations.
"""
import numba as _nb
import numpy as _np
from numba import njit


@njit(nogil=True)  # no parallel jit found
def batched_vector_matrix(b1: _np.ndarray, b2: _np.ndarray, matrix: _np.ndarray):  # pragma: no cover
    """

    Args:
        b1: a numpy array containing all the particles
        b2: explicit destination for the result
        matrix: the transfer matrix as a numpy array

    Returns:
        the destination (result) array
    """
    return b1, _np.dot(b1, matrix.T, out=b2)


@njit(parallel=True, nogil=True)
def batched_vector_tensor(b1: _np.ndarray, b2: _np.ndarray, tensor: _np.ndarray):  # pragma: no cover
    """

    Args:
        b1:
        b2:
        tensor:

    Returns:

    """
    for h in _nb.prange(b1.shape[0]):
        for i in range(tensor.shape[0]):
            s = 0.0
            for j in range(tensor.shape[1]):
                for k in range(j, tensor.shape[2]):  # Assume upper triangular matrix Mjk = Ti::
                    s += tensor[i, j, k] * b1[h, j] * b1[h, k]
            b2[h, i] = s
    return b1, b2


@njit(nogil=True)
def batched_vector_matrix_tensor(
    b1: _np.ndarray,
    b2: _np.ndarray,
    matrix: _np.ndarray,
    tensor: _np.ndarray,
):  # pragma: no cover
    """

    Args:
        b1:
        b2:
        matrix:
        tensor:

    Returns:

    """
    for h in range(b1.shape[0]):
        for i in range(tensor.shape[0]):
            s = 0
            for j in range(tensor.shape[1]):
                s += matrix[i, j] * b1[h, j]
                for k in range(j, tensor.shape[2]):  # Assume upper triangular matrix Mjk = Ti::
                    s += tensor[i, j, k] * b1[h, j] * b1[h, k]
            b2[h, i] = s
    return b1, b2


@njit
def matrix_matrix(m1, m2):  # pragma: no cover
    return _np.matmul(m1, m2)



"""
TODO
"""
from __future__ import annotations

import inspect
import uuid
from typing import Any, Callable, Dict, Optional, Tuple

import numpy as _np
from georges_core.patchable import Patchable as _Patchable
from numba.typed import List as nList
from pint import UndefinedUnitError as _UndefinedUnitError

from ... import ureg as _ureg
from ..apertures import (
    circular_aperture_check,
    elliptical_aperture_check,
    phase_space_aperture_check,
    rectangular_aperture_check,
)
from ..integrators import IntegratorType, MadXIntegrator


class ManzoniException(Exception):
    """Exception raised for errors in the Manzoni elements module."""

    def __init__(self, m):
        self.message = m


class ManzoniAttributeException(ManzoniException):
    """Exception raised for errors in the Manzoni elements module."""

    pass


class ElementType(type):
    """
    Dark magic.
    Be careful.

    TODO
    """

    def __new__(mcs, name: str, bases: Tuple[ElementType, type, ...], dct: Dict[str, Any]):
        # Insert a default initializer (constructor) in case one is not present
        if "__init__" not in dct:

            def default_init(self, name: str = "", integrator: IntegratorType = MadXIntegrator, *params, **kwargs):
                """Default initializer for all Commands."""
                defaults = {}
                if "post_init" in dct:
                    defaults = {
                        parameter_name: parameter_value.default
                        for parameter_name, parameter_value in inspect.signature(dct["post_init"]).parameters.items()
                        if parameter_value.default is not inspect.Parameter.empty
                    }
                bases[0].__init__(self, name, integrator, dct.get("PARAMETERS", {}), *params, **{**defaults, **kwargs})
                if "post_init" in dct:
                    dct["post_init"](self, **kwargs)

            dct["__init__"] = default_init

        # Collect all post_init arguments
        if "_POST_INIT" not in dct:
            dct["_POST_INIT"] = {}
        if "post_init" in dct and len(bases) > 0:
            dct["_POST_INIT"] = {*getattr(bases[0], "_POST_INIT", {}), *dct["post_init"].__code__.co_varnames}

        # Add PARAMETERS from the base class
        try:
            dct["PARAMETERS"] = {**getattr(bases[0], "PARAMETERS", {}), **dct.get("PARAMETERS", {})}
        except IndexError:
            pass

        return super().__new__(mcs, name, bases, dct)

    def __init__(cls, name: str, bases: Tuple[type, ...], dct: Dict[str, Any]):
        super().__init__(name, bases, dct)
        if cls.__doc__ is not None:
            cls.__doc__ = cls.__doc__.rstrip()
            cls.__doc__ += """

    .. rubric:: Command attributes

    Attributes:
            """
            for k, v in cls.PARAMETERS.items():
                if isinstance(v, tuple) and len(v) >= 2:
                    cls.__doc__ += f"""
        {k}='{v[0]}' ({type(v[0]).__name__}): {v[1]}
            """

    def __getattr__(cls, key: str):
        try:
            if key.endswith("_"):
                return cls.PARAMETERS[key.rstrip("_")][2]
            else:
                return cls.PARAMETERS[key][0]
        except KeyError:
            raise AttributeError(key)

    def __getitem__(cls, item: str):
        try:
            return cls.PARAMETERS[item]
        except KeyError:
            raise AttributeError(item)

    def __contains__(cls, item) -> bool:
        return item in cls.PARAMETERS


class Element(metaclass=ElementType):
    """Test test test.

    More info on this wonderful class.
    TODO
    """

    PARAMETERS: dict = {
        "NAME": ("", "Primary label for the Manzoni command (default: auto-generated hash)."),
        "AT_ENTRY": (0 * _ureg.m, "Entrance position of the element."),
        "AT_CENTER": (0 * _ureg.m, "Entrance position of the element."),
        "AT_EXIT": (0 * _ureg.m, "Exit position of the element."),
    }
    """Parameters of the element, with their default value and their description ."""

    def __init__(self, name: str = "", *params, **kwargs):
        """
        TODO
        Args:
            name:
            *params:
            **kwargs:
        """
        self._attributes = {}
        for d in (Element.PARAMETERS,) + params:
            self._attributes = dict(self._attributes, **{k: v[0] for k, v in d.items()})
        for k, v in kwargs.items():
            if k not in self._POST_INIT:
                setattr(self, k, v)
        if name:
            self._attributes["NAME"] = name
        if not self._attributes["NAME"]:
            self.generate_label()
        Element.post_init(self, **kwargs)

    def generate_label(self, prefix: str = ""):
        """

        Args:
            prefix:

        Returns:

        """
        self._attributes["NAME"] = "_".join(
            filter(
                None,
                [
                    prefix,
                    str(uuid.uuid4().hex),
                ],
            ),
        )[:20]
        return self

    def post_init(self, **kwargs):  # -> NoReturn:
        """
        TODO

        Args:
            **kwargs: all arguments from the initializer (constructor) are passed to ``post_init`` as keyword arguments.

        """
        pass

    def __getattr__(self, a: str) -> Any:
        """
        TODO
        Args:
            a:

        Returns:

        """
        if self._attributes.get(a) is None:
            try:
                return super().__getattribute__(a)
            except AttributeError:
                return None
        return self._attributes[a]

    def __setattr__(self, k: str, v: Any):
        """
        Custom attribute setter; all non-protected (starting with a '_') attributes in upper-case are considered
        parameters of the `Command`. As such, the method will verify that they are indeed part of the command
        definition, if not an exception is raised. For valid attributes, their dimensionality is verified against the
        command definition (the dimension of the parameter's default value).

        It is also possible to use unit inference by appending an underscore to the attributes' name. In that
        case the unit of the default value is implicitely used. This is useful in case it is known that the parameter's
        numerical value is expressed in Manzoni's default units set.

        Examples:
            >>> c = Element()
            >>> c.NAME = 'FOOBAR'

        Args:
            k: a string representing the attribute
            v: the attribute's value to be set

        Raises:
            A ManzoniException is raised in case the parameter is not part of the class definition or if it has
            invalid dimension.
        """
        if k.startswith("_") or not k.isupper():
            super().__setattr__(k, v)
        else:
            k_ = k.rstrip("_")
            if k_ not in self._attributes.keys():
                raise ManzoniAttributeException(
                    f"The parameter {k_} is not part of the {self.__class__.__name__} ",
                )

            default = self._retrieve_default_parameter_value(k_)
            try:  # Avoid a bug in pint where a string starting with '#' cannot be parsed
                default = default.lstrip("#")
            except AttributeError:
                pass
            if isinstance(v, (int, float)) and k.endswith("_"):
                v = _ureg.Quantity(v, _ureg.Quantity(default).units)
            elif isinstance(v, str) and default is not None and not isinstance(default, str) and not v.startswith("#"):
                try:
                    v = _ureg.Quantity(v)
                except _UndefinedUnitError:
                    pass
            try:
                dimension = v.dimensionality
            except AttributeError:
                dimension = _ureg.Quantity(1).dimensionality  # No dimension
            try:
                if default is not None and dimension != _ureg.Quantity(default).dimensionality:
                    raise ManzoniAttributeException(
                        f"Invalid dimension ({dimension} "
                        f"instead of {_ureg.Quantity(default).dimensionality}) "
                        f"for parameter {k_}={v} of {self.__class__.__name__}.",
                    )
            except (ValueError, TypeError, _UndefinedUnitError):
                pass
            self._attributes[k_] = v

    def _retrieve_default_parameter_value(self, k: str) -> Any:
        """
        Retrieve the default value of a given parameter as defined in the Command definition (class hierarchy).

        Args:
            k: the parameter for which the default value is requested.

        Returns:
            the default value of the Command's parameter 'k'.
        """
        try:
            return self.PARAMETERS[k][0]
        except (TypeError, IndexError):
            return self.PARAMETERS[k]

    def __repr__(self) -> str:
        return str(self)

    def __str__(self) -> str:
        """
        Provides the string representation of the command in the Manzoni input file format.

        Returns:
            The string representation.

        Examples:
            >>> c = Element('my_label_1')
            >>> str(c) #doctest: +NORMALIZE_WHITESPACE
            "Element: {'NAME': 'my_label_1', 'AT_ENTRY': <Quantity(0, 'meter')>,
            'AT_CENTER': <Quantity(0, 'meter')>, 'AT_EXIT': <Quantity(0, 'meter')>}"

        """
        return f"{self.__class__.__name__}: {str(self.attributes)}"

    @property
    def attributes(self) -> Dict[str, _ureg.Quantity]:
        """All attributes.

        Provides a dictionary with all attributes for the command.

        Returns: dictionnary with all attributes.

        """
        return self._attributes

    @property
    def defaults(self) -> Dict[str, _ureg.Quantity]:
        """Default attributes.

        Provides a dictionary with all attributes that have been assigned a default value.

        Returns: dictionary with all default attributes.
        """
        return {k: v for k, v in self._attributes.items() if v == self.PARAMETERS.get(k)[0]}

    @property
    def nondefaults(self) -> Dict[str, _ureg.Quantity]:
        """Non default attributes.

        Provides a dictionary with all attributes that have been assigned a non default value.

        Returns: dictionary with all non default attributes.
        """
        return {k: v for k, v in self._attributes.items() if v != self.PARAMETERS.get(k)[0]}


class ManzoniElement(Element, _Patchable):
    INTEGRATOR: Optional[IntegratorType] = MadXIntegrator

    def __init__(self, name: str = "", integrator: Optional[IntegratorType] = None, *params, **kwargs):
        """

        Args:
            name:
            integrator:
            *params:
            **kwargs:
        """
        super().__init__(name, *params, **kwargs)
        self._integrator: Optional[IntegratorType] = integrator or self.INTEGRATOR
        self._cache: Optional[nList] = None
        self._frozen: bool = False

    def propagate(
        self,
        beam: _np.ndarray,
        out: Optional[_np.ndarray] = None,
        global_parameters: list = None,
    ) -> Tuple[_np.ndarray, _np.ndarray]:
        """

        Args:
            beam:
            out:
            global_parameters:

        Returns:

        """

        return self.integrator.propagate(self, beam, out, global_parameters)

    def check_aperture(
        self,
        beam: _np.ndarray,
        out: _np.ndarray,
    ):
        """

        Args:
            beam:
            out:

        Returns:

        """
        if self.APERTYPE is not None:
            aperture_check, aperture_parameters = self.aperture
            out = _np.compress(
                aperture_check(out, aperture_parameters),
                out,
                axis=0,
            )
        return beam, out

    def freeze(self):
        """

        Returns:

        """
        self.cache  # Calls it!
        self._frozen = True
        return self

    def unfreeze(self):
        """

        Returns:

        """
        self._frozen = False
        return self

    @property
    def frozen(self):
        """

        Returns:

        """
        return self._frozen

    @property
    def unfrozen(self):
        """

        Returns:

        """
        return not self._frozen

    @property
    def integrator(self) -> IntegratorType:
        """

        Returns:

        """
        return self._integrator

    @integrator.setter
    def integrator(self, integrator: IntegratorType):
        self._integrator = integrator

    @property
    def parameters(self) -> list:
        return []

    @property
    def aperture(self) -> Optional[Tuple[Callable, _np.ndarray]]:
        if self.APERTYPE.upper() == "CIRCULAR":
            return (
                circular_aperture_check,
                _np.array(
                    [
                        self.APERTURE[0].m_as("meter"),
                    ],
                ),
            )
        elif self.APERTYPE.upper() == "RECTANGULAR":
            return (
                rectangular_aperture_check,
                _np.array(
                    [
                        self.APERTURE[0].m_as("meter"),
                        self.APERTURE[1].m_as("meter"),
                    ],
                ),
            )
        elif self.APERTYPE.upper() == "ELLIPTICAL":
            return (
                elliptical_aperture_check,
                _np.array(
                    [
                        self.APERTURE[0].m_as("meter"),
                        self.APERTURE[1].m_as("meter"),
                    ],
                ),
            )
        elif self.APERTYPE.upper() == "PHASE_SPACE":
            return (
                phase_space_aperture_check,
                _np.array(
                    [
                        self.APERTURE[0],
                        self.APERTURE[1],
                        self.APERTURE[2],
                        self.APERTURE[3],
                    ],
                ),
            )
        else:
            return None

    def clear_cache(self):
        self._cache = None

    @property
    def cache(self) -> nList:
        if not self.frozen:
            self._cache = nList()
            for e in self.integrator.cache(self):
                self._cache.append(e)
        return self._cache



from typing import List, Optional

import numpy as _np
import pandas as _pd
from georges_core.distribution import Distribution
from lmfit import Model, Parameters
from numba import njit


class ObserverType(type):
    pass


class Observer(metaclass=ObserverType):
    def __init__(self, elements):
        self.elements = elements
        self.data = []
        self.headers = []

    def __call__(self, element: Optional[List[str]], b1, b2):
        if self.elements is None:
            return True
        if element.NAME not in self.elements:
            return False
        else:
            return True

    def to_df(self) -> _pd.DataFrame:
        return _pd.DataFrame(self.data, columns=self.headers).set_index("NAME")


class BeamObserver(Observer):
    """

    Return the beam distribution at the exit of an element. Optionnaly, return the beam at the entrance of the element.

    """

    def __init__(self, elements: Optional[List[str]] = None, with_input_beams: bool = False):
        super().__init__(elements)
        self._with_input_beams = with_input_beams
        self.headers = (
            "NAME",
            "AT_ENTRY",
            "AT_CENTER",
            "AT_EXIT",
            "BEAM_IN",
            "BEAM_OUT",
        )

    def __call__(self, element, b1, b2):
        if super().__call__(element, b1, b2):
            self.data.append(
                (
                    element.NAME,
                    element.AT_ENTRY,
                    element.AT_CENTER,
                    element.AT_EXIT,
                    _np.copy(b1) if self._with_input_beams else None,
                    _np.copy(b2),
                ),
            )


class SuperObserver(Observer):
    def __init__(self, elements: Optional[List[str]] = None):
        super().__init__(elements)
        self.headers = (
            "NAME",
            "AT_ENTRY",
            "AT_CENTER",
            "AT_EXIT",
        )

    def __call__(self, element, b1, b2):
        if super().__call__(element, b1, b2):
            self.data.append((element.NAME, element.AT_ENTRY, element.AT_CENTER, element.AT_EXIT))


class MeanObserver(Observer):
    """

    Compute the mean values of the beam coordinates.

    """

    def __init__(self, elements: Optional[List[str]] = None):
        super().__init__(elements)
        self.headers = (
            "NAME",
            "AT_ENTRY",
            "AT_CENTER",
            "AT_EXIT",
            "BEAM_IN_X",
            "BEAM_OUT_X",
            "BEAM_IN_Y",
            "BEAM_OUT_Y",
            "BEAM_IN_XP",
            "BEAM_OUT_XP",
            "BEAM_IN_YP",
            "BEAM_OUT_YP",
            "BEAM_IN_DPP",
            "BEAM_OUT_DPP",
        )

    def __call__(self, element, b1, b2):
        if super().__call__(element, b1, b2):
            self.data.append(
                (
                    element.NAME,
                    element.AT_ENTRY,
                    element.AT_CENTER,
                    element.AT_EXIT,
                    b1[:, 0].mean(),
                    b2[:, 0].mean(),
                    b1[:, 2].mean(),
                    b2[:, 2].mean(),
                    b1[:, 1].mean(),
                    b2[:, 1].mean(),
                    b1[:, 3].mean(),
                    b2[:, 3].mean(),
                    b1[:, 4].mean(),
                    b2[:, 4].mean(),
                ),
            )


class SigmaObserver(Observer):
    """

    Compute the standard deviation of the beam coordinates.

    """

    def __init__(self, elements: Optional[List[str]] = None):
        super().__init__(elements)
        self.headers = (
            "NAME",
            "AT_ENTRY",
            "AT_CENTER",
            "AT_EXIT",
            "BEAM_IN_X",
            "BEAM_OUT_X",
            "BEAM_IN_Y",
            "BEAM_OUT_Y",
            "BEAM_IN_XP",
            "BEAM_OUT_XP",
            "BEAM_IN_YP",
            "BEAM_OUT_YP",
            "BEAM_IN_DPP",
            "BEAM_OUT_DPP",
        )

    def __call__(self, element, b1, b2):
        if super().__call__(element, b1, b2):
            self.data.append(
                (
                    element.NAME,
                    element.AT_ENTRY,
                    element.AT_CENTER,
                    element.AT_EXIT,
                    b1[:, 0].std(),
                    b2[:, 0].std(),
                    b1[:, 2].std(),
                    b2[:, 2].std(),
                    b1[:, 1].std(),
                    b2[:, 1].std(),
                    b1[:, 3].std(),
                    b2[:, 3].std(),
                    b1[:, 4].std(),
                    b2[:, 4].std(),
                ),
            )


class LossesObserver(Observer):
    """

    Compute the losses and the transmission in an element.

    """

    def __init__(self, elements: Optional[List[str]] = None):
        super().__init__(elements)
        self.headers = (
            "NAME",
            "AT_ENTRY",
            "AT_CENTER",
            "AT_EXIT",
            "PARTICLES_IN",
            "PARTICLES_OUT",
            "TRANSMISSION",
            "LOSSES",
        )

    def __call__(self, element, b1, b2):
        if super().__call__(element, b1, b2):
            self.data.append(
                (
                    element.NAME,
                    element.AT_ENTRY,
                    element.AT_CENTER,
                    element.AT_EXIT,
                    b1.shape[0],
                    b2.shape[0],
                    100 * (b2.shape[0] / b1.shape[0]),
                    100 * (1 - b2.shape[0] / b1.shape[0]),
                ),
            )

    def compute_global_transmission(self, global_transmission: float = 1.0):
        for elem in self.data:
            global_transmission *= elem[3] / 100
        return global_transmission * 100

    def compute_global_losses(self):
        return 100 - self.compute_global_transmission()

    # def adjust_for_efficiency(self, efficiency: float = 1.0):
    # do something with self.data
    # self.data['LOSSES'] /= efficiency
    # self.data['TRANSMISSION'] *= effiency
    #   ...


class SymmetryObserver(Observer):
    """

    Compute the symmetry of the beam.

    """

    def __init__(self, elements: Optional[List[str]] = None):
        super().__init__(elements=elements)

        self.headers = (
            "NAME",
            "AT_ENTRY",
            "AT_CENTER",
            "AT_EXIT",
            "SYM_IN",
            "SYM_OUT",
        )

    def __call__(self, element, b1, b2):
        self.data.append(
            (
                element.NAME,
                element.AT_ENTRY,
                element.AT_CENTER,
                element.AT_EXIT,
                abs(b1[:, 0].std() - b1[:, 2].std()) / (b1[:, 0].std() + b1[:, 2].std()),
                abs(b2[:, 0].std() - b2[:, 2].std()) / (b2[:, 0].std() + b2[:, 2].std()),
            ),
        )


class TwissObserver(Observer):

    """

    Compute the Twiss parameters of the beam.

    """

    def __init__(self, elements=None):
        super().__init__(elements)
        self.headers = (
            "NAME",
            "AT_ENTRY",
            "AT_CENTER",
            "AT_EXIT",
            "EMIT_IN_X",
            "EMIT_OUT_X",
            "BETA_IN_X",
            "BETA_OUT_X",
            "ALPHA_IN_X",
            "ALPHA_OUT_X",
            "DISP_IN_X",
            "DISP_OUT_X",
            "DISP_IN_XP",
            "DISP_OUT_XP",
            "EMIT_IN_Y",
            "EMIT_OUT_Y",
            "BETA_IN_Y",
            "BETA_OUT_Y",
            "ALPHA_IN_Y",
            "ALPHA_OUT_Y",
            "DISP_IN_Y",
            "DISP_OUT_Y",
            "DISP_IN_YP",
            "DISP_OUT_YP",
        )

    def __call__(self, element, b1, b2):
        if super().__call__(element, b1, b2):
            # Do not use distribution but directly use numpy with numba
            twiss_in = Distribution.compute_twiss(b1)
            twiss_out = Distribution.compute_twiss(b2)

            self.data.append(
                (
                    element.NAME,
                    element.AT_ENTRY,
                    element.AT_CENTER,
                    element.AT_EXIT,
                    twiss_in[0],
                    twiss_out[0],
                    twiss_in[1],
                    twiss_out[1],
                    twiss_in[2],
                    twiss_out[2],
                    twiss_in[3],
                    twiss_out[3],
                    twiss_in[4],
                    twiss_out[4],
                    twiss_in[5],
                    twiss_out[5],
                    twiss_in[6],
                    twiss_out[6],
                    twiss_in[7],
                    twiss_out[7],
                    twiss_in[8],
                    twiss_out[8],
                    twiss_in[9],
                    twiss_out[9],
                ),
            )


class IbaBpmObserver(Observer):
    def __init__(self, elements: Optional[List[str]] = None):
        super().__init__(elements)
        self.headers = (
            "NAME",
            "AT_ENTRY",
            "AT_CENTER",
            "AT_EXIT",
            "BEAM_OUT_X",
            "BEAM_OUT_Y",
        )

    @staticmethod
    def fit_bpm(distribution: _np.array):
        @njit
        def gaussian(x, a, mu, sigma):
            return a * _np.exp(-((x - mu) ** 2) / (2 * sigma**2)) / (_np.sqrt(2 * _np.pi) * sigma)

        def fit_bpm(d, maxfev=1e7):
            bs = (
                _np.array(
                    [
                        -31,
                        -19.8,
                        -15.8,
                        -11.8,
                        -7.8,
                        -5.8,
                        -3.8,
                        -1.8,
                        0.0,
                        1.8,
                        3.8,
                        5.8,
                        7.8,
                        11.8,
                        15.8,
                        19.8,
                        31,
                    ],
                )
                / 1000
            )
            bsp = (bs[1:] + bs[:-1]) / 2
            w = 1.0 / (bs[1:] - bs[:-1])
            w[0] *= 0.7
            w[-1] *= 0.7
            hist = _np.histogram(d, bs)
            x = bsp
            y = w * hist[0]
            ar = _np.trapz(y / _np.sum(y) * len(y), x)
            mean = _np.mean(x * y / _np.sum(y) * len(y))
            rms = _np.std(x * y / _np.sum(y) * len(y))

            params = Parameters()  # Instanciate the Parameters class, then add variables as keywords
            params.add("a", value=ar, min=-1e6, max=1e6)
            params.add("mu", value=mean, min=mean - _np.abs(mean), max=mean + _np.abs(mean))
            params.add("sigma", value=rms, min=0.5 * rms, max=2 * rms)
            params.add("max_nfev", value=maxfev)

            gmodel = Model(gaussian)
            result = gmodel.fit(
                data=y,
                x=x,
                params=params,
            )

            # Print the fit result
            print("bpm fit result:", result.best_values["a"], result.best_values["mu"], result.best_values["sigma"])
            return [result.best_values["sigma"]]

        return fit_bpm(distribution)

    def __call__(self, element, b1, b2):
        if super().__call__(element, b1, b2) and element.CLASS == "Marker":
            print(element.NAME)  # To identify the BPM whose data are being fitted
            self.data.append(
                (
                    element.NAME,
                    element.AT_ENTRY,
                    element.AT_CENTER,
                    element.AT_EXIT,
                    self.fit_bpm(b2[:, 0])[0],
                    self.fit_bpm(b2[:, 2])[0],
                ),
            )



"""
The `elements` submodule contains the physical implementation of the different beamline elements,
namely:

- Drifts
- Dipoles
- Quadrupoles
- Sextupoles
- Multipoles (magnets with a combined dipolar, quadrupolar and sextupolar function),
- Collimators (rectangular, circular and elliptical);
- Scatterers
    thin materials that interact with the particles and only modify the transverse angle along the two transverse axes
- Degraders
    thick materials which combine particle interaction with a simple drift type propagation along the  material
"""
from .collimators import CircularCollimator, Collimator, Dump, EllipticalCollimator, RectangularCollimator
from .elements import ManzoniElement
from .magnets import (
    Bend,
    DipEdge,
    Drift,
    Fringein,
    Fringeout,
    Gap,
    HKicker,
    Kicker,
    Marker,
    Matrix,
    Multipole,
    Quadrupole,
    RBend,
    SBend,
    Sextupole,
    SRotation,
    TKicker,
    VKicker,
)
from .scatterers import BeamStop, Degrader, Scatterer



"""
Plotly plotting module for Manzoni.
"""

from __future__ import annotations

import logging
from typing import Union

import cpymad.madx
import numpy as _np
import pandas as _pd
from georges_core.vis import PlotlyArtist as _PlotlyArtist
from georges_core.vis.artist import PALETTE

from ..manzoni.observers import BeamObserver as _BeamObserver
from ..manzoni.observers import LossesObserver as _LossesObserver
from ..manzoni.observers import MeanObserver as _MeanObserver
from ..manzoni.observers import Observer as _Observer
from ..manzoni.observers import SigmaObserver as _SigmaObserver
from ..manzoni.observers import TwissObserver as _TwissObserver

palette = PALETTE["solarized"]
palette["both"] = palette["base03"]
palette["X"] = palette["cyan"]
palette["Y"] = palette["orange"]
palette["XP"] = palette["red"]
palette["YP"] = palette["green"]


class BeamPlottingException(Exception):
    """Exception raised for errors in the beam plotting module."""

    def __init__(self, m):
        self.message = m


class ManzoniPlotlyArtist(_PlotlyArtist):
    """A plotly implementation of a `Plotly` artist."""

    def __init__(self, tracks_color: str = "b", **kwargs):
        """
        Args:
            param ax: the matplotlib ax used for plotting. If None it will be created with `init_plot` (kwargs are
            forwarded).
            tracks_color: color for the plotting of tracks
            kwargs: forwarded to `ZgoubiPlot` and to `init_plot`.
        """
        super().__init__(**kwargs)
        self._tracks_color = tracks_color

    def tracking(
        self,
        observer: _Observer = None,
        plane: str = "X",
        fill_between: bool = False,
        mean: bool = True,
        std: bool = False,
        halo: bool = True,
        **kwargs,
    ):
        """Plot the beam envelopes from tracking data.


        Args:
            observer (_Observer, optional): Observer used for the tracking. Defaults to None.
            plane (str, optional): Plane to draw. Defaults to "X".
            fill_between (bool, optional): Fill the gap between the lines in transparency. Defaults to False.
            mean (bool, optional): Add the mean to the plot. Defaults to True.
            std (bool, optional): Add the std to the plot. Defaults to False.
            halo (bool, optional): Add a halo to the plot. Defaults to True.

        """
        tracking_palette = kwargs.get("palette", palette)
        df_observer = observer.to_df()

        if isinstance(observer, _MeanObserver):
            self.scatter(
                x=_np.hstack([0, df_observer["AT_EXIT"].apply(lambda e: e.m_as("m")).values]),
                y=_np.hstack(
                    [
                        df_observer.iloc[0][f"BEAM_IN_{plane}"],
                        df_observer.iloc[:][f"BEAM_OUT_{plane}"].values,
                    ],
                )
                * 1000,
                marker={"symbol": 4, "color": tracking_palette[plane], "size": 7},
                line={"width": 1},
                mode="markers+lines",
                showlegend=False,
                name=kwargs.get("name", plane),
            )
            self.layout["xaxis"]["title"] = "S (m)"
            self.layout["yaxis"]["title"] = "Mean position (mm)"
        elif isinstance(observer, _SigmaObserver):
            x = _np.hstack([0, df_observer["AT_EXIT"].apply(lambda e: e.m_as("m")).values])

            if plane == "both":
                y0 = (
                    _np.hstack(
                        [
                            df_observer.iloc[0]["BEAM_IN_X"],
                            df_observer.iloc[:]["BEAM_OUT_X"].values,
                        ],
                    )
                    * 1000
                )
                y1 = (
                    _np.hstack(
                        [
                            df_observer.iloc[0]["BEAM_IN_Y"],
                            df_observer.iloc[:]["BEAM_OUT_Y"].values,
                        ],
                    )
                    * 1000
                )
            else:
                y0 = (
                    _np.hstack(
                        [
                            df_observer.iloc[0][f"BEAM_IN_{plane}"],
                            df_observer.iloc[:][f"BEAM_OUT_{plane}"].values,
                        ],
                    )
                    * 1000
                )
                y1 = y0
            y = [y1, -y0]

            self.scatter(
                x=x,
                y=y[0],
                marker={"symbol": 4, "color": tracking_palette[plane], "size": 7},
                line={"width": 1, "color": tracking_palette[plane]},
                mode="markers+lines",
                name=kwargs.get("name", plane),
                showlegend=False,
            )

            self.scatter(
                x=x,
                y=y[1],
                marker={"symbol": 4, "color": tracking_palette[plane], "size": 7},
                line={"width": 1, "color": tracking_palette[plane]},
                mode="markers+lines",
                showlegend=False,
                name=kwargs.get("name", plane),
                fill="tonexty" if fill_between else None,
                fillcolor=tracking_palette[plane],
            )

            self.layout["xaxis"]["title"] = "S (m)"
            self.layout["yaxis"]["title"] = "Beam Size (mm)"

            if plane == "both":
                self.fig["data"][0]["line"]["color"] = tracking_palette["Y"]
                self.fig["data"][0]["marker"]["color"] = tracking_palette["Y"]
                self.fig["data"][1]["line"]["color"] = tracking_palette["X"]
                self.fig["data"][1]["marker"]["color"] = tracking_palette["X"]
                ticks_val = _np.arange(_np.min(10 * _np.floor(y[1] / 10)), _np.max(10 * _np.ceil(y[0] / 10)) + 10, 10)
                self.layout["yaxis"]["tickvals"] = ticks_val
                self.layout["yaxis"]["ticktext"] = [_np.abs(d) for d in ticks_val]

                # Optimized for a general layout
                self.layout["yaxis"]["title"] = "<--------------   Beam size (mm)   -------------->"
                self.layout["annotations"] = [
                    {
                        "x": -0.075,
                        "y": 0.65 + self.layout["height"] / 6000,
                        "xref": "paper",
                        "yref": "paper",
                        "text": "Vertical",
                        "textangle": -90,
                    },
                    {
                        "x": -0.075,
                        "y": self.layout["height"] / 6000,
                        "xref": "paper",
                        "yref": "paper",
                        "text": "Horizontal",
                        "textangle": -90,
                    },
                ]

        elif isinstance(observer, _BeamObserver):
            dico_plane = {"X": 0, "PX": 1, "Y": 2, "PY": 3}
            t = df_observer.apply(
                lambda r: _pd.Series(
                    {
                        "S": r["AT_EXIT"].m_as("m"),
                        "mean": 1000 * r["BEAM_OUT"][:, dico_plane[plane]].mean() if mean else 0.0,
                        "std": 1000 * r["BEAM_OUT"][:, dico_plane[plane]].std() if std else 0.0,
                        "1%": 1000
                        * (
                            self.compute_halo(r["BEAM_OUT"][:, dico_plane[plane]], 0.023)
                            - self.compute_halo(r["BEAM_OUT"][:, dico_plane[plane]], 0.5)
                        ),
                        "5%": 1000
                        * (
                            self.compute_halo(r["BEAM_OUT"][:, dico_plane[plane]], 0.159)
                            - self.compute_halo(r["BEAM_OUT"][:, dico_plane[plane]], 0.5)
                        ),
                        "95%": 1000
                        * (
                            self.compute_halo(r["BEAM_OUT"][:, dico_plane[plane]], 0.841)
                            - self.compute_halo(r["BEAM_OUT"][:, dico_plane[plane]], 0.5)
                        ),
                        "99%": 1000
                        * (
                            self.compute_halo(r["BEAM_OUT"][:, dico_plane[plane]], 0.977)
                            - self.compute_halo(r["BEAM_OUT"][:, dico_plane[plane]], 0.5)
                        ),
                    },
                ),
                axis=1,
            )

            if df_observer.iloc[0]["BEAM_IN"] is not None:
                data_entry = df_observer.iloc[0]
                t0 = _pd.DataFrame(
                    data={
                        "S": [data_entry["AT_ENTRY"].m_as("m")],
                        "mean": [
                            1000
                            * data_entry["BEAM_IN"][
                                :,
                                dico_plane[plane],
                            ].mean()
                            if mean
                            else 0.0,
                        ],
                        "std": [
                            1000 * data_entry["BEAM_IN"][:, dico_plane[plane]].std() if std else 0.0,
                        ],
                        "1%": [
                            1000
                            * (
                                self.compute_halo(
                                    data_entry["BEAM_IN"][:, dico_plane[plane]],
                                    0.023,
                                )
                                - self.compute_halo(
                                    data_entry["BEAM_IN"][:, dico_plane[plane]],
                                    0.5,
                                )
                            ),
                        ],
                        "5%": [
                            1000
                            * (
                                self.compute_halo(
                                    data_entry["BEAM_IN"][:, dico_plane[plane]],
                                    0.159,
                                )
                                - self.compute_halo(
                                    data_entry["BEAM_IN"][:, dico_plane[plane]],
                                    0.5,
                                )
                            ),
                        ],
                        "95%": [
                            1000
                            * (
                                self.compute_halo(
                                    data_entry["BEAM_IN"][:, dico_plane[plane]],
                                    0.841,
                                )
                                - self.compute_halo(
                                    data_entry["BEAM_IN"][:, dico_plane[plane]],
                                    0.5,
                                )
                            ),
                        ],
                        "99%": [
                            1000
                            * (
                                self.compute_halo(
                                    data_entry["BEAM_IN"][:, dico_plane[plane]],
                                    0.977,
                                )
                                - self.compute_halo(
                                    data_entry["BEAM_IN"][:, dico_plane[plane]],
                                    0.5,
                                )
                            ),
                        ],
                    },
                    index=["Start"],
                )
                t = _pd.concat([t0, t])

            if t["S"].count == 0:
                return

            if halo:
                self.filled_plot(
                    self,
                    t["S"],
                    t["mean"] + t["5%"],
                    t["mean"] + t["95%"],
                    tracking_palette[plane],
                    fill=True,
                    alpha=0.3,
                )
                self.filled_plot(
                    self,
                    t["S"],
                    t["mean"] + t["1%"],
                    t["mean"] + t["99%"],
                    tracking_palette[plane],
                    fill=True,
                    alpha=0.3,
                )
            #
            if mean:
                self.scatter(
                    x=t["S"],
                    y=t["mean"],
                    marker={"symbol": 2, "color": tracking_palette[plane], "size": 7},
                    line={"width": 1},
                    mode="markers+lines",
                    showlegend=False,
                    name="mean",
                )

            if std:
                self.scatter(
                    x=t["S"],
                    y=t["mean"] + t["std"],
                    marker={"symbol": 2, "color": tracking_palette[plane], "size": 7},
                    line={"width": 1},
                    mode="markers+lines",
                    showlegend=False,
                    name="std",
                )

                self.scatter(
                    x=t["S"],
                    y=t["mean"] - t["std"],
                    marker={"symbol": 2, "color": tracking_palette[plane], "size": 7},
                    line={"width": 1},
                    mode="markers+lines",
                    showlegend=False,
                    name="std",
                )

            self.layout["xaxis"]["title"] = "S (m)"
            self.layout["yaxis"]["title"] = "Beam Size (mm)"

        elif isinstance(observer, _LossesObserver):
            raise BeamPlottingException("Use method vis.ManzoniPlotlyArtist().losses to plot losses.")

        else:
            raise BeamPlottingException(f"No plotting method for {observer} is implemented")

    def losses(self, observer: _LossesObserver = None, log_scale: bool = False, **kwargs):
        """
        Plot the losses along the beamline
        Args:
            observer: Observer used for the tracking
            log_scale: Log scale for transmission
        """

        if not isinstance(observer, _LossesObserver):
            raise BeamPlottingException("The observer must be a LossesObserver.")

        losses_palette = kwargs.get("palette", palette)
        df_observer = observer.to_df()
        exit_pos = df_observer["AT_EXIT"].apply(lambda e: e.m_as("m"))

        self.layout["xaxis"]["title"] = "S (m)"
        self.layout["yaxis"]["title"] = r"Losses (%)"
        self.layout["yaxis"]["titlefont"] = {"color": losses_palette["magenta"]}

        self.bar(
            x=exit_pos,
            y=df_observer["LOSSES"],
            marker={"color": losses_palette["magenta"]},
            width=0.125,
            name="Losses",
            showlegend=False,
        )

        max_val = (df_observer["LOSSES"]).abs().max()
        self.layout["yaxis"]["dtick"] = _np.ceil(max_val / 10)
        self.layout["yaxis"]["range"] = [0, max_val + 5.0]

        self.add_secondary_axis(title=r"T (%)")
        self.layout["yaxis2"]["titlefont"] = {"color": losses_palette["green"]}

        # TODO USE Transmission, shift and compute ?
        init = df_observer.iloc[0]["PARTICLES_IN"]
        global_transmission = 100 * (df_observer["PARTICLES_OUT"].values / init)

        if log_scale:
            self.scatter(
                x=_np.hstack([0, exit_pos.values]),
                y=_np.hstack([100, global_transmission]),
                mode="lines+markers",
                marker={"symbol": 17, "color": losses_palette["green"]},
                yaxis="y2",
                name="Transmission",
                showlegend=False,
            )
            self.layout["yaxis2"]["type"] = "log"
            self.layout["yaxis2"]["range"] = [_np.floor(_np.log(_np.min(global_transmission)) / 10) - 1, 2]
            yticksval = 10 ** (_np.arange(_np.floor(_np.log(_np.min(global_transmission) / 10)) - 1, 3, 1))
            self.layout["yaxis2"]["tickvals"] = yticksval

        else:
            self.scatter(
                x=_np.hstack([0, exit_pos.values]),
                y=_np.hstack([100, global_transmission]),
                yaxis="y2",
                mode="lines+markers",
                marker={"symbol": 1, "color": losses_palette["green"]},
                name="Transmission",
                showlegend=False,
            )
            self.layout["yaxis2"]["dtick"] = 10
            self.layout["yaxis2"]["range"] = [0, 100]

    @staticmethod
    def compute_halo(data, percentile):
        """Return a dataframe containing the 1st, 5th, 95th and 99th percentiles of each dimensions."""
        return _np.quantile(data, percentile)

    @staticmethod
    def filled_plot(ax, x, y0, y, c, fill=False, **kwargs):
        ax.scatter(x=x, y=y, mode="lines", line={"width": 1, "color": c}, showlegend=False)
        if fill:
            ax.scatter(x=x, y=y0, mode="lines", line={"width": 1, "color": c}, showlegend=False, fill="tonexty")

    def twiss(
        self,
        observer: _TwissObserver = None,
        with_beta: bool = True,
        with_alpha: bool = False,
        with_dispersion: bool = False,
        tfs_data: Union[_pd.DataFrame, cpymad.madx.Table] = None,
        relativistic_beta: float = 1.0,
        **kwargs,
    ):
        """Plot the Twiss function along the beamline

        Args:
            observer (_TwissObserver, optional): TwissObserver. Defaults to None.
            with_beta (bool, optional): Add beta function to the plot. Defaults to True.
            with_alpha (bool, optional): Add the alpha function ot the plot. Defaults to False.
            with_dispersion (bool, optional): Add the dispersion to the plot. Defaults to False.
            tfs_data (Union[_pd.DataFrame, cpymad.madx.Table], optional): tfs file from MAD. Defaults to None.
            relativistic_beta (float): Relativistic beta value to scale the dispersion. Default to 1.
        """
        tracking_palette = kwargs.get("palette", palette)
        df_observer = observer.to_df()

        if isinstance(tfs_data, cpymad.madx.Table):
            tfs_data = _pd.DataFrame(
                data={
                    "S": tfs_data.s,
                    "BETX": tfs_data.betx,
                    "BETY": tfs_data.bety,
                    "ALFX": tfs_data.alfx,
                    "ALFY": tfs_data.alfy,
                    "DX": tfs_data.dx,
                    "DY": tfs_data.dy,
                },
            )

        if with_beta:
            self.scatter(
                x=_np.hstack([0, df_observer["AT_EXIT"].apply(lambda e: e.m_as("m")).values]),
                y=_np.hstack(
                    [
                        df_observer.iloc[0]["BETA_IN_X"],
                        df_observer.iloc[:]["BETA_OUT_X"].values,
                    ],
                ),
                line={"width": 1, "color": tracking_palette["blue"]},
                mode="lines",
                showlegend=True,
                name=r"$\beta_x \text{ - Manzoni}$",
            )
            self.scatter(
                x=_np.hstack([0, df_observer["AT_EXIT"].apply(lambda e: e.m_as("m")).values]),
                y=_np.hstack(
                    [
                        df_observer.iloc[0]["BETA_IN_Y"],
                        df_observer.iloc[:]["BETA_OUT_Y"].values,
                    ],
                ),
                line={"width": 1, "color": tracking_palette["red"]},
                mode="lines",
                showlegend=True,
                name=r"$\beta_y \text{ - Manzoni}$",
            )
            self.layout["yaxis"]["title"] = r"$\beta \text{(m)}$"

            if tfs_data is not None:
                self.scatter(
                    x=tfs_data["S"].values,
                    y=tfs_data["BETX"].values,
                    marker={"symbol": 4, "color": tracking_palette["blue"], "size": 7},
                    mode="markers",
                    showlegend=True,
                    name=r"$\beta_x \text{ - MAD-X}$",
                )

                self.scatter(
                    x=tfs_data["S"].values,
                    y=tfs_data["BETY"].values,
                    marker={"symbol": 4, "color": tracking_palette["red"], "size": 7},
                    mode="markers",
                    showlegend=True,
                    name=r"$\beta_y \text{ - MAD-X}$",
                )
        if with_alpha:
            self.scatter(
                x=_np.hstack([0, df_observer["AT_EXIT"].apply(lambda e: e.m_as("m")).values]),
                y=_np.hstack(
                    [
                        df_observer.iloc[0]["ALPHA_IN_X"],
                        df_observer.iloc[:]["ALPHA_OUT_X"].values,
                    ],
                ),
                marker={"symbol": 4, "color": tracking_palette["blue"], "size": 7},
                line={"width": 1},
                mode="markers+lines",
                showlegend=True,
                name=r"$\alpha_x \text{- Manzoni}$",
            )
            self.scatter(
                x=_np.hstack([0, df_observer["AT_EXIT"].apply(lambda e: e.m_as("m")).values]),
                y=_np.hstack(
                    [
                        df_observer.iloc[0]["ALPHA_IN_Y"],
                        df_observer.iloc[:]["ALPHA_OUT_Y"].values,
                    ],
                ),
                marker={"symbol": 4, "color": tracking_palette["red"], "size": 7},
                line={"width": 1},
                mode="markers+lines",
                showlegend=True,
                name=r"$\alpha_y$ \text{- Manzoni}$",
            )
            self.layout["yaxis"]["title"] = r"$\alpha$"

            if tfs_data is not None:
                self.scatter(
                    x=tfs_data["S"].values,
                    y=tfs_data["ALFX"].values,
                    marker={"symbol": 4, "color": tracking_palette["blue"], "size": 7},
                    mode="markers",
                    showlegend=True,
                    name=r"$\alpha_x \text{- MAD-X}$",
                )

                self.scatter(
                    x=tfs_data["S"].values,
                    y=tfs_data["ALFY"].values,
                    marker={"symbol": 4, "color": tracking_palette["red"], "size": 7},
                    mode="markers",
                    showlegend=True,
                    name=r"$\alpha_y \text{- MAD-X}$",
                )
        if with_dispersion:
            self.add_secondary_axis("Dispersion")
            self.scatter(
                x=_np.hstack([0, df_observer["AT_EXIT"].apply(lambda e: e.m_as("m")).values]),
                y=_np.hstack(
                    [
                        df_observer.iloc[0]["DISP_IN_X"],
                        df_observer.iloc[:]["DISP_OUT_X"].values,
                    ],
                ),
                marker={"symbol": 4, "color": tracking_palette["blue"], "size": 7},
                line={"width": 1},
                mode="markers+lines",
                showlegend=True,
                yaxis="y2",
                name=r"$D_x \text{- Manzoni}$",
            )
            self.scatter(
                x=_np.hstack([0, df_observer["AT_EXIT"].apply(lambda e: e.m_as("m")).values]),
                y=_np.hstack(
                    [
                        df_observer.iloc[0]["DISP_IN_Y"],
                        df_observer.iloc[:]["DISP_OUT_Y"].values,
                    ],
                ),
                marker={"symbol": 4, "color": tracking_palette["red"], "size": 7},
                line={"width": 1},
                mode="markers+lines",
                showlegend=True,
                yaxis="y2",
                name=r"$D_y \text{- Manzoni}$",
            )
            self.layout["yaxis2"]["title"] = r"Dispersion"

            if tfs_data is not None:
                logging.warning(
                    f"Dispersion from MAD-X is multiplied by the beta relativistic factor {relativistic_beta}.",
                )
                self.scatter(
                    x=tfs_data["S"].values,
                    y=tfs_data["DX"].values * relativistic_beta,
                    marker={"symbol": 4, "color": tracking_palette["blue"], "size": 7},
                    mode="markers",
                    showlegend=True,
                    yaxis="y2",
                    name=r"$D_x \text{- MAD-X}$",
                )

                self.scatter(
                    x=tfs_data["S"].values,
                    y=tfs_data["DY"].values * relativistic_beta,
                    marker={"symbol": 4, "color": tracking_palette["red"], "size": 7},
                    mode="markers",
                    showlegend=True,
                    yaxis="y2",
                    name=r"$D_y \text{- MAD-X}$",
                )
        self.layout["xaxis"]["title"] = "S (m)"

    def phase_space(
        self,
        observer: _BeamObserver = None,
        element: str = None,
        location: str = "OUT",
        dim=None,
        nbins=None,
        draw_ellipse: bool = True,
    ):
        # TODO
        logging.error("The method is not yet implemented")



"""
The file `apertures.py` is for the `collimator` elements and contains the implementation of the
selection of the particle according to the type of collimator (circular, elliptical, rectangular or phase-space).
"""
import numpy as _np
from numba import njit


@njit
def circular_aperture_check(b1, kargs: _np.ndarray):
    """
    The aperture is circular and defined by its radius.

    Args:
        b1: The beam before the collimator
        kargs: Radius of the aperture.

    Returns:
        The collimated beam

    """
    return (b1[:, 0] ** 2 + b1[:, 2] ** 2) < kargs[0] ** 2


@njit
def elliptical_aperture_check(b1, kargs: _np.ndarray):
    """
    The aperture is elliptical and defined by the semi-axes of the elliptical aperture.

    Args:
        b1: The beam before the collimator
        kargs: Semi-axes of the aperture.

    Returns:
        The collimated beam

    """
    return (b1[:, 0] / kargs[0]) ** 2 + (b1[:, 2] / kargs[1] ** 2) < 1


@njit
def rectangular_aperture_check(b1, kargs: _np.ndarray):
    """
    The aperture is rectangular and defined by its horizontal and vertical half apertures.

    Args:
        b1: The beam before the collimator
        kargs: horizontal and vertical half apertures

    Returns:
        The collimated beam

    """
    return _np.multiply(_np.abs(b1[:, 0]) < kargs[0], _np.abs(b1[:, 2]) < kargs[1])


@njit
def phase_space_aperture_check(b1, kargs: _np.ndarray):
    """
    The phase space aperture allows the user to cut either the position or the momentum of the beam.

    Args:
        b1: The beam before the collimator
        kargs: Radius of the aperture in each coordinates system (position and momentum).

    Returns:
        The collimated beam

    """
    return ((b1[:, 0] / kargs[0]) ** 2 + (b1[:, 1] / kargs[1]) ** 2) < 1 and (
        (b1[:, 2] / kargs[2]) ** 2 + (b1[:, 3] / kargs[3]) ** 2
    ) < 1



"""
The `maps` submodule contains the physics of the propagation of the particles through each of the
beamline elements described in the `elements` submodule. For each element, a first and second-order
type propagation is implemented, allowing the user to select the order of the tracking that is suitable
for his specific application. It is done via the selection of the integrator type when building the
beamline (see later). The `Transport-type`, `MadX-type`, and `Mad8-type` maps are implemented and
available. The user should be aware that the canonical variables of the particles are not the same
for the three different integrator types, so the definition of the beam must be done according to
the integrator to be consistent.
"""

from .mad8_combined_dipole import compute_mad_combined_dipole_matrix, compute_mad_combined_dipole_tensor
from .mad8_drift import compute_mad_drift_matrix, compute_mad_drift_tensor, drift6
from .mad8_quadrupole import compute_mad_quadrupole_matrix, compute_mad_quadrupole_tensor
from .madx_combined_dipole import tmsect
from .madx_thick import (
    track_madx_bend,
    track_madx_dipedge,
    track_madx_drift,
    track_madx_drift_paraxial,
    track_madx_kicker,
    track_madx_quadrupole,
    track_madx_srotation,
)
from .transport_combined_dipole import (
    compute_transport_combined_dipole_matrix,
    compute_transport_combined_dipole_tensor,
)
from .transport_combined_dipole_ex import (
    compute_transport_combined_dipole_ex_matrix,
    compute_transport_combined_dipole_ex_tensor,
)
from .transport_drift import compute_transport_drift_matrix
from .transport_fringe_in import compute_transport_fringe_in_matrix, compute_transport_fringe_in_tensor
from .transport_fringe_in_ex import compute_transport_fringe_in_ex_matrix, compute_transport_fringe_in_ex_tensor
from .transport_fringe_out import compute_transport_fringe_out_matrix, compute_transport_fringe_out_tensor
from .transport_fringe_out_ex import compute_transport_fringe_out_ex_matrix, compute_transport_fringe_out_ex_tensor
from .transport_multipole import compute_transport_multipole_matrix, compute_transport_multipole_tensor
from .transport_multipole_ex import compute_transport_multipole_ex_matrix, compute_transport_multipole_ex_tensor
from .transport_quadrupole import compute_transport_quadrupole_matrix, compute_transport_quadrupole_tensor
from .transport_quadrupole_ex import compute_transport_quadrupole_ex_matrix, compute_transport_quadrupole_ex_tensor
from .transport_sextupole import compute_transport_sextupole_matrix, compute_transport_sextupole_tensor
from .transport_sextupole_ex import compute_transport_sextupole_ex_matrix, compute_transport_sextupole_ex_tensor
